from Song import Song, SongLibrary
from Track import Track
from dynamic_markov_chain import DynamicMarkovChain, chainType

# Create a new song object and read in the desired training song
training_song = Song()
training_song.load(filename=SongLibrary.NIRVANA_LITHIUM)

# Create a chord and note markov chain with the given attributes
chord_chain = DynamicMarkovChain("chord chain", token_length=3, chain_type=chainType.CHORD)
note_chain = DynamicMarkovChain("note chain", token_length=4, chain_type=chainType.NOTE)

# Train the new markov chains with the desired song
chord_chain.add_song(training_song)
note_chain.add_song(training_song)

# Create a new song to write to
output_song = Song()

# Create verse sections from the chord and note chains
chd_verse = chord_chain.generate_pattern(output_song, 4, 46, octave=4)
mel_verse = note_chain.generate_pattern(output_song, 16, 72, octave=5)

# Create chorus sections from the chord and note chains
chd_chorus = chord_chain.generate_pattern(output_song, 4, 46, octave=4)
mel_chorus = note_chain.generate_pattern(output_song, 8, 24, octave=5)
mel_chorus2 = note_chain.generate_pattern(output_song, 8, 24, octave=5)

# Create an optional counter melody with the note chain
# bass_mel = note_chain.generate_pattern(output_song, 16, 18, octave=4)

# Create output tracks (different to the individual sections created earlier)
chord_track = Track(channel=2)
melody_track = Track(channel=0)
melody_track_2 = Track(channel=1)

# "Write" the song by stitching together the individual sections into a complete track
chord_track = chord_track.append_tracks([0, 0, 1, 1, 0, 0, 1, 1, 0, 0], [chd_verse, chd_chorus])
melody_track = melody_track.append_tracks([0, 0, 1, 1, 2, 1, 0, 0, 1, 2, 1, 2, 0, 0], [mel_verse, mel_chorus, mel_chorus2])
# melody_track_2 = melody_track_2.append_tracks([0, 0, 0, 0, 0, 0, 0, 0], [bass_mel])

# Add these tracks to the song object
output_song.add_track(chord_track)
output_song.add_track(melody_track)
output_song.add_track(melody_track_2)

# Save the new written song
output_song.save(filename='../../MIDI Files/Demo Output/Structured Song8.mid')


from Genre import Genre
from Song import Song
import os
from Track import TagEnum
from Key import Key
from markov_chain import MarkovChain, Type
from markov_library import MarkovLibrary
from dynamic_markov_chain import DynamicMarkovChain, chainType

if __name__ == '__main__':

    directory = "C:\\Users\\Eli\\Documents\\GitHub\\2021FallTeam17-DeHaan\\MIDI Files\\Rock"
    genre = Genre()
    for artist in os.listdir(directory):
        artist_directory = directory + '\\' + artist.title()
        for song in os.listdir(artist_directory):
            song_object = Song()
            print(song)
            try:
                song_object.load(filename=artist_directory + '\\' + song.title())
            except (IOError, AttributeError) as e:
                continue
            current_key = song_object.detect_key_and_scale()[0]
    #         # print(current_key.tonic)
    #         # song_object.change_song_key(current_key, Key())
            genre.add_song(song_object)

    chord_chain = DynamicMarkovChain("chord chain", token_length=3, chain_type=chainType.CHORD)
    note_chain = DynamicMarkovChain("note chain", token_length=3, chain_type=chainType.NOTE)
    for song in genre.songs:
        try:
            chord_chain.add_song(song)
        except AttributeError as e:
            continue
        note_chain.add_song(song)

    song2 = Song()
    chord_track = chord_chain.generate_pattern(song2, 25, 46)
    melody_track = note_chain.generate_pattern(song2, 100, 25)

    song2.add_track(chord_track)
    song2.add_track(melody_track)

    song2.save('generated_song.mid', True)
    #
    # genre.get_notes_frequency_graph()
    # genre.print_songs()

    # song = Song()
    # song.load(filename="../MIDI Files/Rock/Elton John/TinyDancer.mid", print_file=False)
    # chord_chain = DynamicMarkovChain("chord chain", token_length=3, chain_type=chainType.CHORD)
    # note_chain = DynamicMarkovChain("note chain", token_length=3, chain_type=chainType.NOTE)
    # chord_chain.add_song(song)
    # note_chain.add_song(song)
    # song2 = Song()
    # chord_chain.generate_pattern(song2, 25, 46)
    # note_chain.generate_pattern(song2, 100, 25)
    # song2.save('generated_song.mid', True)


    # song.load(filename="../MIDI Files/Hip-Hop/Kanye West/24851_Gold-Digger.mid", print_file=True)
    # song.load(filename="../test/test MIDI/C_major_chords.mid")
    # song.load(filename="../MIDI Files/Utility/C_Major_Pentatonic.mid", print_file=False)
    # song.load(filename="../MIDI Files/Pop/Adele/Someone_Like_You_easy_piano.mid", print_file=False)
    # song.change_song_key(Key(tonic=song.detect_key_and_scale()[0:1]), Key())
    # song.get_transition_graph(name="Someone Like You")
    # song.load(filename="../MIDI Files/Rock/Elton John/TinyDancer.mid", print_file=False)
    # song.change_song_key(Key(tonic=song.detect_key_and_scale()[0:1]), Key())
    # song.get_transition_graph(name="Tiny Dancer")
    # song.get_transition_graph(name="Skyfall")
    # song.change_song_key(Key(tonic=song.detect_key_and_scale()[0:1]), Key())
    # song.get_transition_graph(name="Your Song")

    # for track in song.tracks:
    #     print("track: " + track.track_name)
    #     for chord in track.chords:
    #         print("  Chord: ")
    #         for note in chord.notes:
    #             print("    " + str(note.pitch) + " Time: " + str(note.time) + " channel: " + str(note.channel))

    # print(song.to_string())

    # song.print_song()

    # song.load(filename="../MIDI Files/Hip-Hop/Kanye West/24851_Gold-Digger.mid", print_file=False)
    # song.get_note_frequency_graph("Gold Digger by Kanye West")
    # song.get_note_velocity_graph("Gold Digger by Kanye West")
    # song.load(filename="music samples/Mii Channel.mid", print_file=True)
    # song.load(filename="music samples/Megadeth-Symphony Of Destruction.mid", print_file=True)

    # print(song.to_string())

    # print(song.detect_key())

    # for track in song.tracks:
    #     print(track.track_name + " -- " + str(track.tag))

    # song.detect_key_by_phrase_endings()

    # for track in song.tracks:

    # song.change_song_key(origin_key=Key('F#', 'major'), destination_key=Key('C', 'major'))
    # song.save(filename="music samples/Megadeth-Tornado of Souls.mid", print_file=True)

    # song.save(filename="music samples/Mii Channel Output.mid", print_file=True)



"""
    These array values are semitone differences from the previous note. This is formatted this way because
    notes in the scale change based on the tonic.

    If this is confusing here is an example:

    For the Major scale, the notes in the scale include the tonic (or Key).
    Then, how you know what other notes are in the scale you need to make hops of a specified amount of semitones
    (half steps [1], whole steps [2], minor third [3], etc.) away from the previous note in the scale. So,
    C Major is defined as C (tonic), D (a whole step away from previous), E (a whole steep away from D),
    F (only a half step away from E) and so on...
"""

SCALE_TYPES = dict(
    major=[2, 2, 1, 2, 2, 2, 1],
    dorian=[2, 1, 2, 2, 2, 1, 2],
    phrygian=[1, 2, 2, 2, 1, 2, 2],
    lydian=[2, 2, 2, 1, 2, 2, 1],
    mixolydian=[2, 2, 1, 2, 2, 1, 2],
    minor=[2, 1, 2, 2, 1, 2, 2],
    locrian=[1, 2, 2, 1, 2, 2, 2],
)

"""
    other potential scales/modes that may be useful later
    
    pentatonic=[2, 2, 3, 2],
    pentatonic_minor=[3, 2, 2, 3],
    blues=[3, 2, 1, 1, 3, 2],
    dorian=[2, 1, 2, 2, 2, 1, 2],
    phrygian=[1, 2, 2, 2, 1, 2, 2],
    lydian=[2, 2, 2, 1, 2, 2, 1],
    mixolydian=[2, 2, 1, 2, 2, 1, 2],
    aeolian=[2, 1, 2, 2, 1, 2, 2],
    locrian=[1, 2, 2, 1, 2, 2, 2],
    harmonic_major=[2, 2, 1, 2, 1, 3, 1],
    harmonic_minor=[2, 1, 2, 2, 1, 3, 1],
    super_locrian=[1, 2, 1, 2, 2, 2, 2],
    japanese=[1, 4, 2, 3, 2],
    akebono=[2, 1, 4, 2, 3],
"""



"""
    This is the example for how to use the chord transition graph functionality. It creates a directed
    cyclical graph with all the chords in the song and the transition that occurs between them. The edges
    depict the number of times that transition occured in the song.

"""

from Song import Song

# Create an empty song object
song = Song()

# Load the MIDI file into the song object
song.load(filename="../../MIDI Files/Rock/Elton John/RocketMan.mid")

# Run the method to get the transition graph and view it
song.get_transition_graph(name="Rocket Man")

# The graph will be viewable in a pdf file in the file directory.



from Song import Song
from math import floor
from random import randrange

WHOLE = 1
HALF = WHOLE / 2
QUARTER = HALF / 2
EIGHTH = QUARTER / 2
SIXTEENTH = EIGHTH / 2
THIRTY_SECOND = SIXTEENTH / 2
HALF_TRIPLET = WHOLE / 3
QUARTER_TRIPLET = HALF / 3
EIGHTH_TRIPLET = QUARTER / 3
SIXTEENTH_TRIPLET = EIGHTH / 3
THIRTY_SECOND_TRIPLET = SIXTEENTH / 3
DOTTED_WHOLE = WHOLE + WHOLE / 2
DOTTED_HALF = HALF + HALF / 2
DOTTED_QUARTER = QUARTER + QUARTER / 2
DOTTED_EIGHTH = EIGHTH + EIGHTH / 2
DOTTED_SIXTEENTH = SIXTEENTH + SIXTEENTH/2
DOTTED_THIRTY_SECOND = THIRTY_SECOND + THIRTY_SECOND/2
WHOLE_REST = -WHOLE
HALF_REST = -HALF
QUARTER_REST = -QUARTER
EIGHTH_REST = -EIGHTH
SIXTEENTH_REST = -SIXTEENTH
THIRTY_SECOND_REST = -THIRTY_SECOND


def apply_rhythm_pattern(song=None, track=None, pattern_array=None):
    current_abs_time = track.notes[0].time
    whole_length = floor(song.ticks_per_beat * 4)
    j = 0

    for i, note in enumerate(track.notes):

        pattern_idx = j % len(pattern_array)
        # check for rest
        if pattern_array[pattern_idx] >= 0:
            if i + 1 < len(track.notes):
                next_note_time = track.notes[i + 1].time
            #check for chords
            if note.time != next_note_time:
                note.duration = floor(pattern_array[pattern_idx] * whole_length)
                note.time = floor(current_abs_time)
                current_abs_time += note.duration
                j += 1
            else:
                note.duration = floor(pattern_array[pattern_idx] * whole_length)
                note.time = floor(current_abs_time)
        # Handle rests
        else:
            current_abs_time += -(floor(pattern_array[pattern_idx] * whole_length))
            j += 1
            pattern_idx = j % len(pattern_array)
            if note.time != next_note_time:
                note.duration = floor(pattern_array[pattern_idx] * whole_length)
                note.time = floor(current_abs_time)
                current_abs_time += note.duration
                j += 1
            else:
                note.duration = floor(pattern_array[pattern_idx] * whole_length)
                note.time = floor(current_abs_time)





def humanify_rhythm(song=None, track=None, humanify_percent=0.5):

    max_error = 0.12
    if song is None or track is None or humanify_percent > 1 or humanify_percent < 0:
        raise SyntaxError("Song and track must not be None. and humanify_percent needs to be"
                          "between 0.0 and 1.0")

    max_time_offset = floor(song.ticks_per_beat * max_error * humanify_percent)
    for i, note in enumerate(track.notes):
        #check for chord
        if i + 1 < len(track.notes):
            next_note_time = track.notes[i + 1].time

            new_time_offset = note.time + randrange(-max_time_offset, max_time_offset)
            new_duration_offset = note.duration + randrange(-max_time_offset, max_time_offset, 1)

            if new_time_offset >= 0 and new_duration_offset >= 0:
                note.time = new_time_offset
                note.duration = new_duration_offset




from Song import Song, SongLibrary
from Key import Key, Mode

# Create an empty song object
song = Song()

# Load the MIDI file into the song object
song.load(filename=SongLibrary.AMAZING_GRACE)

# Manually enter the key to change from/to
song.change_song_key(Key("G", Mode.MAJOR), Key("G", Mode.MINOR))

# Use the auto-detected key of the song
# song.change_song_key(song.key, Key("G", Mode.MINOR))


song.save(filename="../../MIDI Files/Demo Output/Amazing_Grace_Major.mid")



from BatchReporting import generate_batch_key_detection_report

# Using this method you can generate a CSV file which will run key detection on an
# entire file directory formatted with sub directories as genre/artist/song.mid
# This CSV file will be created in a new directory (under the same PWD as the calling
# script)
generate_batch_key_detection_report(top_directory_path="../../MIDI Files")



from Song import Song, SongLibrary
from Track import Track
from dynamic_markov_chain import DynamicMarkovChain, chainType

# Create a new song object and read in the desired training song
training_song = Song()
training_song.load(filename=SongLibrary.NIRVANA_LITHIUM)

# Create a chord and note markov chain with the given attributes
chord_chain = DynamicMarkovChain("chord chain", token_length=3, chain_type=chainType.CHORD)
note_chain = DynamicMarkovChain("note chain", token_length=4, chain_type=chainType.NOTE)

# Train the new markov chains with the desired song
chord_chain.add_song(training_song)
note_chain.add_song(training_song)

# Create a new song to write to
output_song = Song()

# Create a chord and melody track
chords = chord_chain.generate_pattern(output_song, num_notes=16, instrument=46, octave=4)
melody = note_chain.generate_pattern(output_song, num_notes=64, instrument=72, octave=5)

# Add the generated tracks to a new song
output_song.add_track(chords)
output_song.add_track(melody)

# Save the new written song
output_song.save(filename='../../MIDI Files/Demo Output/Basic_Markov.mid')



import logging

import Note
from enum import Enum

# The channel percussion information will be on
PERCUSSION_CHANNEL = 9
# The average note pitch on a track must be below this number to be considered a bass track
BASS_AVERAGE = 45
# The chord to note ratio required to classify a track as "chords"
CHORD_PERCENTAGE = 0.15


class Track:

    def __init__(self, notes=None, controls=None, track_name=None, device_name=None, chords=None, channel=0):
        """ Constructor for the Track object

        Args:
            notes (Note[], optional): List of Notes within this track. Defaults to None.
            controls (Control[], optional): List of Control messages within this track. Defaults to None.
            track_name (String, optional): Name of this track. Defaults to None.
            device_name (String, optional): Instrument used on this track. Defaults to None.
            channel (int, optional): Channel of this track. Defaults to 0.
        """        
        if notes is None:
            notes = []
        if chords is None:
            chords = []
        # Control messages (and program change messages) store data about how the track should be played back,
        # but are not necessarily "events" in the file. These can occur at any point during a song
        if controls is None:
            controls = []
        self.notes = notes
        self.controls = controls
        self.track_name = track_name
        self.device_name = device_name
        self.channel = channel
        self.chords = chords
        self.tag = TagEnum.NONE

        if channel is PERCUSSION_CHANNEL:
            self.is_percussion = True
        else:
            self.is_percussion = False

    def add_note(self, note=None):
        """ Adds a note to the end of this track

        Args:
            note (Note): Note to append to the song. Defaults to None.
        """        
        self.notes.append(note)

    def add_chord(self, chord=None):
        """ Adds a chord to the track array

        Args:
            chord (Chord): Note to append to the song. Defaults to None.
        """
        self.chords.append(chord)

    def get_c_indexed_note_frequencies(self):
        """ Returns an array representing each note's number of appearances in this track, starting
        with C.

        Returns:
            int[]: Note frequencies array
        """        
        c_indexed_frequencies = [0] * Note.NUM_NOTES

        for note in self.notes:
            c_indexed_frequencies[note.c_indexed_pitch_class] += 1

        return c_indexed_frequencies

    def generate_tags(self):
        """ Sets 'tag' field in the song to the appropriate tag based on its attributes
        This method assumes a track is only used for one section of a song and does not
        dramatically change roles during the song (ex, switch from guitar track to vocal
        track)

        """
        if len(self.notes) == 0:
            self.tag = TagEnum.NONE

        elif self.channel is PERCUSSION_CHANNEL:
            self.tag = TagEnum.PERCUSSION

        else:
            pitch_total = 0
            for note in self.notes:
                pitch_total += note.pitch
            if pitch_total / len(self.notes) < BASS_AVERAGE:
                self.tag = TagEnum.BASS      # If the average note pitch is lower than BASS_AVERAGE
            elif len(self.chords) > CHORD_PERCENTAGE * len(self.notes):
                self.tag = TagEnum.CHORDS
            else:
                self.tag = TagEnum.MELODY    # If nothing else fits, this is likely a melody track

    def get_all_chords(self):
        return self.chords

    def get_unique_chords(self):
        chord_set = set()
        return [x for x in self.chords if x not in chord_set and not chord_set.add(x)]

    def append_track(self, track):
        """
        "Glues" two tracks together. Returns a new track with the contents of the two given tracks. Does not modify
        either track.

        :param track: The track to attach to the end of this track
        :return: A new track with the contents of the two tracks played consecutively.
        """

        if track is None:
            raise AttributeError("Track must not be None")

        last_event_time = 0
        for note in self.notes:
            if note.time + note.duration > last_event_time:
                last_event_time = note.time + note.duration
        for control in self.controls:
            if control.time > last_event_time:
                last_event_time = control.time

        new_self = self.duplicate_track()
        new_track = track.duplicate_track()

        for note in new_track.notes:
            if not note.chord_note:
                note.time += last_event_time
                new_self.notes.append(note)

        for control in new_track.controls:
            control.time += last_event_time
            new_self.controls.append(control)

        for chord in new_track.chords:
            chord.time += last_event_time
            new_self.chords.append(chord)

            for note in chord.notes:
                note.time += last_event_time
                new_self.notes.append(note)

        return new_self

    def append_tracks(self, pattern=None, sections=None):
        """Takes a pattern as an array of ints (Ex. [0, 0, 1, 1, 2, 1, 0]) and strings together
        tracks (supplied in sections) in the pattern given. For example, if the given pattern is [0, 1, 2, 1, 0] and
        the given sections are in the format [a, b, c], this will return a new track with the structure (a b c b a).

        Params:
        pattern: An array of ints used to dictate the pattern of tracks to return. The highest value must be less
            than the length of "sections"
        sections: A list of tracks that will be glued together in this method. The order given will determine how
            the pattern is constructed (ie. the index of each section in the list corresponds to the int value in
            "pattern" that will be used to represent it)

        Returns:
            A new track with the given sections attached to it in the given pattern.
        """
        new_track = self.duplicate_track()
        for i in range(len(pattern)):
            new_track = new_track.append_track(sections[pattern[i]])

        return new_track

    def duplicate_track(self):
        """
        Returns a new track with identical contents to the given track. Does not modify the given track.
        :return: a new track with identical contents to the given track.
        """
        notes = []
        controls = []
        chords = []

        for note in self.notes:
            if not note.chord_note:
                notes.append(note.duplicate_note())

        for control in self.controls:
            controls.append(control.duplicate_control())

        for chord in self.chords:
            chords.append(chord.duplicate_chord())
            for note in chord.notes:
                notes.append(note.duplicate_note())

        return Track(notes=notes, controls=controls, track_name=self.track_name, device_name=self.device_name,
                     chords=chords, channel=self.channel)

    def equals(self, track):
        """
        Returns true if these two tracks are equal. If they are not equal, prints an info log message with more details
        and return false
        :param track: The track to compare this track to
        :return: True, if the contents of the track are equal. False otherwise.
        """
        if self.channel != track.channel:
            logging.info(msg="tracks " + track.track_name + " have different channel values")
            logging.info(msg="This: " + self.channel + ", compare to: " + track.channel)
            return False

        if self.track_name != track.track_name:
            logging.info(msg="tracks " + track.track_name + " have different track names :")
            logging.info(msg="This: '" + self.track_name + "', compare to: '" + track.track_name + "'")
            return False

        if self.device_name != track.device_name:
            logging.info(msg="tracks " + track.track_name + " have different device names")
            logging.info(msg="This: '" + self.device_name + "', compare to: '" + track.device_name + "'")
            return False

        if len(self.controls) != len(track.controls):
            logging.info(msg="tracks " + track.track_name + " have different numbers of control messages")
            logging.info(msg="This: " + str(len(self.controls)) + ", compare to: "
                             + str(len(track.controls)))
            return False

        for j, control in enumerate(self.controls):
            if control.msg_type != track.controls[j].msg_type:
                logging.info(msg="tracks " + track.track_name + " control " + str(j) + " have different message types")
                logging.info(msg="This: '" + control.msg_type + "', compare to: '"
                                 + track.controls[j].msg_type + "'")
                return False

            if control.control != track.controls[j].control:
                logging.info(msg="tracks " + track.track_name + " control " + str(j) + " have different control numbers")
                logging.info(msg="This: " + str(control.control) + ", compare to: "
                                 + str(track.controls[j].control))
                return False

            if control.value != track.controls[j].value:
                logging.info(msg="tracks " + track.track_name + " control " + str(j) + " have different values")
                logging.info(msg="This: " + str(control.value) + ", compare to: "
                                 + str(track.controls[j].value))
                return False

            if control.tempo != track.controls[j].tempo:
                logging.info(msg="tracks " + track.track_name + " control " + str(j) + " have different tempos")
                logging.info(msg="This: " + str(control.tempo) + ", compare to: "
                                 + str(track.controls[j].tempo))
                return False

            if control.instrument != track.controls[j].instrument:
                logging.info(msg="tracks " + track.track_name + " control " + str(j) + " have different instruments")
                logging.info(msg="This: " + str(control.instrument) + ", compare to: " + str(
                    track.controls[j].instrument))
                return False

            if control.time != track.controls[j].time:
                logging.info(msg="tracks " + track.track_name + " control " + str(j) + " have different times")
                logging.info(msg="This: " + str(control.time) + ", compare to: "
                                 + str(track.controls[j].time))
                return False

        if len(self.notes) != len(track.notes):
            logging.info(msg="tracks " + track.track_name + " have different numbers of notes")
            logging.info(msg="This: " + str(len(track.notes)) + ", compare to: " + str(len(track.notes)))
            return False

        for j, note in enumerate(self.notes):
            if note.pitch != track.notes[j].pitch:
                logging.info(msg="tracks " + track.track_name + " note " + str(j) + " have different pitches")
                logging.info(msg="This: " + str(note.pitch) + ", compare to: " + str(track.notes[j].pitch))
                return False

            if note.time != track.notes[j].time:
                logging.info(msg="tracks " + track.track_name + " note " + str(j) + " have different times")
                logging.info(msg="This: " + str(note.time) + ", compare to: " + str(track.notes[j].time))
                return False

            if note.duration != track.notes[j].duration:
                logging.info(msg="tracks " + track.track_name + " note " + str(j) + " have different durations")
                logging.info(msg="This: " + str(note.duration) + ", compare to: "
                                 + str(track.notes[j].duration))
                return False

            if note.velocity != track.notes[j].velocity:
                logging.info(msg="tracks " + track.track_name + " note " + str(j) + " have different velocities")
                logging.info(msg="This: " + str(note.velocity) + ", compare to: "
                                 + str(track.notes[j].velocity))
                return False

            if note.channel != track.notes[j].channel:
                logging.info(msg="tracks " + track.track_name + " note " + str(j) + " have different channels")
                logging.info(msg="This: " + str(note.channel) + ", compare to: "
                                 + str(track.notes[j].channel))
                return False
        return True


class TagEnum(Enum):
    NONE = 0
    MELODY = 1
    CHORDS = 2
    BASS = 3
    PERCUSSION = 4





import pytest
import FileIO as FileIO
from Control import Control
from Note import Note
from Song import Song
from Track import Track
from Key import Key, Mode
import mido


def test_song_constructor():
    """
        Tests the constructor of Song
    """    
    tracks = [Track()]

    song = Song()
    assert song.ticks_per_beat == 48
    assert len(song.tracks) == 0
    song = Song(ticks_per_beat=2)
    assert song.ticks_per_beat == 2
    song = Song(tracks=tracks, ticks_per_beat=3)
    assert song.tracks[0] == tracks[0]
    with pytest.raises(ValueError):
        song = Song(ticks_per_beat=-1)


def test_change_song_key_by_half_steps():
    """
        Tests the functionality of changing song key by a
        given number of half steps

        Tests using a C scale changed to a D scale
    """
    orig = Song()
    new_song = Song()
    d_scale = Song()

    orig.load(filename="test MIDI/C_major_scale.mid")
    orig = orig.change_song_key_by_half_steps(2)
    orig.save(filename="test MIDI/C_major_scale_Output.mid")

    new_song.load(filename="test MIDI/C_major_scale_Output.mid")
    d_scale.load(filename="test MIDI/D_major_scale.mid")

    assert new_song.tracks[1].notes[0].pitch == d_scale.tracks[1].notes[0].pitch
    assert new_song.tracks[1].notes[1].pitch == d_scale.tracks[1].notes[1].pitch
    assert new_song.tracks[1].notes[2].pitch == d_scale.tracks[1].notes[2].pitch
    assert new_song.tracks[1].notes[3].pitch == d_scale.tracks[1].notes[3].pitch


def test_change_song_key():
    """
        Tests the functionality of changing song key from one key to another
        Tests using a C scale changed to a D scale
    """
    orig = Song()
    new_song = Song()
    d_scale = Song()

    orig.load(filename="test MIDI/C_major_scale.mid")
    orig.change_song_key(origin_key=Key('C', Mode.MAJOR), destination_key=Key('D', Mode.MAJOR))
    orig.save(filename="test MIDI/C_major_scale_Output.mid")

    new_song.load(filename="test MIDI/C_major_scale_Output.mid")
    d_scale.load(filename="test MIDI/D_major_scale.mid")

    assert new_song.tracks[1].notes[0].pitch == d_scale.tracks[1].notes[0].pitch
    assert new_song.tracks[1].notes[1].pitch == d_scale.tracks[1].notes[1].pitch
    assert new_song.tracks[1].notes[2].pitch == d_scale.tracks[1].notes[2].pitch
    assert new_song.tracks[1].notes[3].pitch == d_scale.tracks[1].notes[3].pitch

    orig.load(filename="test MIDI/C_major_scale.mid")
    orig.change_song_key(origin_key=Key('C', Mode.MAJOR), destination_key=Key('C', Mode.DORIAN))
    assert orig.tracks[1].notes[0].pitch == 48
    assert orig.tracks[1].notes[1].pitch == 50
    assert orig.tracks[1].notes[2].pitch == 51
    assert orig.tracks[1].notes[3].pitch == 53
    assert orig.tracks[1].notes[4].pitch == 55
    assert orig.tracks[1].notes[5].pitch == 57
    assert orig.tracks[1].notes[6].pitch == 58
    assert orig.tracks[1].notes[7].pitch == 60

    orig.load(filename="test MIDI/C_major_scale.mid")
    orig.change_song_key(origin_key=Key('C', Mode.MAJOR), destination_key=Key('C', Mode.PHRYGIAN))
    assert orig.tracks[1].notes[0].pitch == 48
    assert orig.tracks[1].notes[1].pitch == 49
    assert orig.tracks[1].notes[2].pitch == 51
    assert orig.tracks[1].notes[3].pitch == 53
    assert orig.tracks[1].notes[4].pitch == 55
    assert orig.tracks[1].notes[5].pitch == 56
    assert orig.tracks[1].notes[6].pitch == 58
    assert orig.tracks[1].notes[7].pitch == 60

    orig.load(filename="test MIDI/C_major_scale.mid")
    orig.change_song_key(origin_key=Key('C', Mode.MAJOR), destination_key=Key('C', Mode.LYDIAN))
    assert orig.tracks[1].notes[0].pitch == 48
    assert orig.tracks[1].notes[1].pitch == 50
    assert orig.tracks[1].notes[2].pitch == 52
    assert orig.tracks[1].notes[3].pitch == 54
    assert orig.tracks[1].notes[4].pitch == 55
    assert orig.tracks[1].notes[5].pitch == 57
    assert orig.tracks[1].notes[6].pitch == 59
    assert orig.tracks[1].notes[7].pitch == 60

    orig.load(filename="test MIDI/C_major_scale.mid")
    orig.change_song_key(origin_key=Key('C', Mode.MAJOR), destination_key=Key('C', Mode.MIXOLYDIAN))
    assert orig.tracks[1].notes[0].pitch == 48
    assert orig.tracks[1].notes[1].pitch == 50
    assert orig.tracks[1].notes[2].pitch == 52
    assert orig.tracks[1].notes[3].pitch == 53
    assert orig.tracks[1].notes[4].pitch == 55
    assert orig.tracks[1].notes[5].pitch == 57
    assert orig.tracks[1].notes[6].pitch == 58
    assert orig.tracks[1].notes[7].pitch == 60

    orig.load(filename="test MIDI/C_major_scale.mid")
    orig.change_song_key(origin_key=Key('C', Mode.MAJOR), destination_key=Key('C', Mode.MINOR))
    assert orig.tracks[1].notes[0].pitch == 48
    assert orig.tracks[1].notes[1].pitch == 50
    assert orig.tracks[1].notes[2].pitch == 51
    assert orig.tracks[1].notes[3].pitch == 53
    assert orig.tracks[1].notes[4].pitch == 55
    assert orig.tracks[1].notes[5].pitch == 56
    assert orig.tracks[1].notes[6].pitch == 58
    assert orig.tracks[1].notes[7].pitch == 60

    orig.load(filename="test MIDI/C_major_scale.mid")
    orig.change_song_key(origin_key=Key('C', Mode.MAJOR), destination_key=Key('C', Mode.LOCRIAN))
    assert orig.tracks[1].notes[0].pitch == 48
    assert orig.tracks[1].notes[1].pitch == 49
    assert orig.tracks[1].notes[2].pitch == 51
    assert orig.tracks[1].notes[3].pitch == 53
    assert orig.tracks[1].notes[4].pitch == 54
    assert orig.tracks[1].notes[5].pitch == 56
    assert orig.tracks[1].notes[6].pitch == 58
    assert orig.tracks[1].notes[7].pitch == 60

    # This only changes the second half of the scale to phrygian
    orig.load(filename="test MIDI/C_major_scale.mid")
    orig.change_song_key(origin_key=Key('C', Mode.MAJOR), destination_key=Key('C', Mode.PHRYGIAN), interval_begin=300)
    assert orig.tracks[1].notes[0].pitch == 48
    assert orig.tracks[1].notes[1].pitch == 50
    assert orig.tracks[1].notes[2].pitch == 52
    assert orig.tracks[1].notes[3].pitch == 53
    assert orig.tracks[1].notes[4].pitch == 55
    assert orig.tracks[1].notes[5].pitch == 56
    assert orig.tracks[1].notes[6].pitch == 58
    assert orig.tracks[1].notes[7].pitch == 60

def test_transition_graph():
    song = Song()
    song.load(filename="test MIDI/C_major_chords.mid")


def test_equals():
    """
    Tests the equals method in Song that compares two song objects
    """
    song1 = Song(ticks_per_beat=100)   # Standard song
    song2 = Song(ticks_per_beat=100)   # Same as 1
    song3 = Song(ticks_per_beat=100)   # Different notes
    song4 = Song(ticks_per_beat=100)   # Different number of notes
    song5 = Song(ticks_per_beat=100)   # Different Control messages
    song6 = Song(ticks_per_beat=100)   # Different number of control messages
    song7 = Song(ticks_per_beat=100)   # Different track name
    song8 = Song(ticks_per_beat=100)   # Different track device
    song9 = Song(ticks_per_beat=100)   # Different number of tracks
    song10 = Song(ticks_per_beat=200)  # Different ticks_per_beat

    notes1 = []  # Standard set of notes
    notes2 = []  # Different values inside the notes
    notes3 = []  # Different number of notes

    control1 = []  # Standard set of controls
    control2 = []  # Different values inside controls
    control3 = []  # Different number of controls

    notes1.append(Note(pitch=100, time=10, duration=10, velocity=50))
    notes1.append(Note(pitch=110, time=15, duration=15, velocity=55))

    notes2.append(Note(pitch=100, time=10, duration=10, velocity=50))
    notes2.append(Note(pitch=110, time=16, duration=15, velocity=55))

    notes3.append(Note(pitch=100, time=10, duration=10, velocity=50))
    notes3.append(Note(pitch=110, time=15, duration=15, velocity=55))
    notes3.append(Note(pitch=110, time=16, duration=15, velocity=55))

    control1.append(Control(msg_type="program_change", instrument=10, time=100))
    control1.append(Control(msg_type="set_tempo", tempo=500000, time=100))

    control2.append(Control(msg_type="program_change", instrument=10, time=100))
    control2.append(Control(msg_type="set_tempo", tempo=500500, time=100))

    control3.append(Control(msg_type="program_change", instrument=10, time=100))
    control3.append(Control(msg_type="set_tempo", tempo=500000, time=100))
    control3.append(Control(msg_type="control_change", control=100, value=10, time=100))

    song1.add_track(Track(track_name="track1", device_name="Device 1", channel=1, notes=notes1, controls=control1))
    song2.add_track(Track(track_name="track1", device_name="Device 1", channel=1, notes=notes1, controls=control1))
    song3.add_track(Track(track_name="track1", device_name="Device 1", channel=1, notes=notes2, controls=control1))
    song4.add_track(Track(track_name="track1", device_name="Device 1", channel=1, notes=notes3, controls=control1))
    song5.add_track(Track(track_name="track1", device_name="Device 1", channel=1, notes=notes1, controls=control2))
    song6.add_track(Track(track_name="track1", device_name="Device 1", channel=1, notes=notes1, controls=control3))
    song7.add_track(Track(track_name="track2", device_name="Device 1", channel=1, notes=notes1, controls=control1))
    song8.add_track(Track(track_name="track1", device_name="Device 2", channel=1, notes=notes1, controls=control1))
    song9.add_track(Track(track_name="track1", device_name="Device 1", channel=1, notes=notes1, controls=control1))
    song9.add_track(Track(track_name="track2", device_name="Device 2", channel=2, notes=notes1, controls=control1))
    song10.add_track(Track(track_name="track1", device_name="Device 1", channel=1, notes=notes1, controls=control1))

    assert song1.equals(song2)
    assert song1.equals(song3) is False
    assert song1.equals(song4) is False
    assert song1.equals(song5) is False
    assert song1.equals(song6) is False
    assert song1.equals(song7) is False
    assert song1.equals(song8) is False
    assert song1.equals(song9) is False
    assert song1.equals(song10) is False
    assert song2.equals(song1)
    assert song3.equals(song1) is False
    assert song4.equals(song1) is False
    assert song5.equals(song1) is False
    assert song6.equals(song1) is False
    assert song7.equals(song1) is False
    assert song8.equals(song1) is False
    assert song9.equals(song1) is False
    assert song10.equals(song1) is False


def test_detect_key_by_phrase_endings():
    d_mix = Song()
    d_mix_track = Track()
    d_mix_track.notes.append(Note(pitch=62, time=0, duration=100))
    d_mix_track.notes.append(Note(pitch=64, time=100, duration=100))
    d_mix_track.notes.append(Note(pitch=66, time=200, duration=100))
    d_mix_track.notes.append(Note(pitch=67, time=300, duration=100))
    d_mix_track.notes.append(Note(pitch=69, time=400, duration=100))
    d_mix_track.notes.append(Note(pitch=71, time=500, duration=100))
    d_mix_track.notes.append(Note(pitch=72, time=600, duration=100))
    d_mix_track.notes.append(Note(pitch=74, time=700, duration=100))

    d_mix_track.generate_tags()

    d_mix.tracks.append(d_mix_track)

    assert d_mix.detect_key_by_phrase_endings()[0].tonic == "D"
    assert d_mix.detect_key_by_phrase_endings()[0].mode == Mode.MIXOLYDIAN

    song2 = Song()
    tonic_not_in_scale = Track()
    tonic_not_in_scale.notes.append(Note(pitch=60, time=0, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=62, time=100, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=64, time=200, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=65, time=300, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=67, time=400, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=69, time=500, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=71, time=600, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=60, time=700, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=62, time=800, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=64, time=900, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=65, time=1000, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=67, time=1100, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=69, time=1200, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=71, time=1300, duration=100))
    tonic_not_in_scale.notes.append(Note(pitch=61, time=1400, duration=100))

    tonic_not_in_scale.generate_tags()

    song2.tracks.append(tonic_not_in_scale)

    assert song2.detect_key_by_phrase_endings()[0].tonic == "C"
    assert song2.detect_key_by_phrase_endings()[0].mode == Mode.MAJOR





import logging
from logging import info
from Track import Track, TagEnum
from Key import Key, KEYS
from Scale import SCALE_TYPES
from Note import NUM_NOTES
import FileIO as FileIO

import matplotlib.pyplot as plt
import collections
import graphviz

DEFAULT_TICKS_PER_BEAT = 48


class SongLibrary:
    GARTH_BROOKS_TO_MAKE_YOU_FEEL_MY_LOVE = "../../MIDI Files/Country/Garth Brooks/26553_To-Make-You-Feel-My-Love.mid"
    JOHN_ANDERSON_I_WISH_I_COULD_HAVE_BEEN_THERE = "../../MIDI Files/Country/John Anderson/2601_I-Wish-I-Could-Have-Been-There.mid"
    TIM_MCGRAW_SOMETHING_LIKE_THAT = "../../MIDI Files/Country/Tim McGraw/2610_Something-Like-That.mid"
    AMAZING_GRACE = "../../MIDI Files/Gospel/John Newton/Amazing_Grace.mid"
    NIRVANA_LITHIUM = "../../MIDI Files/Grunge/Nirvana/Lithium.mid"
    PEARL_JAM_BETTER_MAN = "../../MIDI Files/Grunge/Pearl Jam/BetterMan.mid"
    SOUNDGARDEN_BLACK_HOLE_SUN = "../../MIDI Files/Grunge/Soundgarden/BlackHoleSun.mid"
    STONE_TEMPLE_PILOTS_CREEP = "../../MIDI Files/Grunge/Stone Temple Pilots/Creep.mid"
    SIMON_AND_GARFUNKEL_SCARBOROUGH_FAIR = "../../MIDI Files/Indie/Simon and Garfunkel/scarborough_fair.mid"
    SIMON_AND_GARFUNKEL_SOUND_OF_SILENCE = "../../MIDI Files/Indie/Simon and Garfunkel/Sound_Of_Silence.mid"
    FRANK_SINATRA_MY_WAY = "../../MIDI Files/Jazz/Frank Sinatra/my_way.mid"
    BLACK_SABBATH_IRON_MAN = "../../MIDI Files/Metal/Black Sabbath/Black Sabbath-Iron Man.mid"
    JUDAS_PRIEST_NIGHT_CRAWLER = "../../MIDI Files/Metal/Judas Priest/Judas Priest - Night Crawler.mid"
    MEGADETH_SYMPHONY_OF_DESTRUCTION = "../../MIDI Files/Metal/Megadeth/Megadeth-Symphony Of Destruction.mid"
    METALLICA_ENTER_SANDMAN = "../../MIDI Files/Metal/Metallica/EnterSandman.mid"
    ADELE_SKYFALL = "../../MIDI Files/Pop/Adele/Adele_-_Skyfall.mid"
    BILLIE_EILISH_NO_TIME_TO_DIE = "../../MIDI Files/Pop/Billie Eilish/Billie Eilish - No Time To Die (James Bond) (midi by Carlo Prato) (www.mid"
    BRUNO_MARS_THE_LAZY_SONG = "../../MIDI Files/Pop/Bruno Mars/Thelazysong.mid"
    ED_SHEERAN_SHAPE_OF_YOU = "../../MIDI Files/Pop/Ed Sheeran/Ed Sheeran - Shape of You  (midi by Carlo Prato) (www.cprato.com).mid"
    IMAGINE_DRAGONS_RADIOACTIVE = "../../MIDI Files/Pop/Imagine Dragons/Radioactive.mid"
    THE_CHAINSMOKERS_FT_HALSEY_CLOSER = "../../MIDI Files/Pop/The Chainsmokers/The Chainsmokers ft. Halsey - Closer  (midi by Carlo Prato) (www.cprato.com).mid"
    THE_WEEKND_PARTY_MONSTER = "../../MIDI Files/Pop/The Weeknd/The Weeknd - Party Monster  (midi by Carlo Prato) (www.cprato.com).mid"
    AEROSMITH_DREAM_ON = "../../MIDI Files/Rock/Aerosmith/DreamOn.mid"
    EAGLES_HOTEL_CALIFORNIA = "../../MIDI Files/Rock/Eagles/HotelCalifornia.mid"
    ELTON_JOHN_ROCKET_MAN = "../../MIDI Files/Rock/Elton John/RocketMan.mid"
    FALL_OUT_BOY_LIGHT_EM_UP = "../../MIDI Files/Rock/Fall Out Boy/LightEmUp.mid"
    GREENDAY_BOULEVARD_OF_BROKEN_DREAMS = "../../MIDI Files/Rock/Greenday/BoulevardofBrokenDreams.mid"
    ROLLING_STONES_SATISFACTION = "../../MIDI Files/Rock/Rolling Stones/Satisfaction.mid"


# Stores metadata about a song, and the tracks included in the song
# <jmleeder>
class Song:

    def __init__(self, tracks=None, ticks_per_beat=DEFAULT_TICKS_PER_BEAT, key=None):
        """ Constructor for the Song class.

        Args:
            tracks (Track[], optional): List of Tracks that make up the song. Defaults to None.
            ticks_per_beat (int, optional): The amount of ticks that pass within one beat in the 
                song. Defaults to DEFAULT_TICKS_PER_BEAT (48).

        Raises:
            ValueError: For a negative ticks_per_beat value
        """
        if tracks is None:
            tracks = []
        self.tracks = tracks
        if ticks_per_beat >= 0:
            self.ticks_per_beat = ticks_per_beat
        else:
            raise ValueError
        if isinstance(key, Key):
            self.key = key
        else:
            self.key = None

    def add_track(self, t):
        """ Adds a new track to the song.

        Args:
            t (Track): Track to add to the song
        """
        assert isinstance(t, Track)
        self.tracks.append(t)

    @staticmethod
    def get_notes_array():
        """ TODO: ?

        Returns:
            String[]: List of each key possible represented as strings.
        """
        return KEYS

    def save(self, filename, print_file=False):
        """ Saves a song object to a midi file with the given name

        Args:
            filename (String): Name of file to save the song as
            print_file (bool, optional): Whether or not to print out the song. Mainly 
                used for debugging purposes. Defaults to False.
        """
        FileIO.write_midi_file(self, filename=filename, print_file=print_file)

    def load(self, filename, print_file=False):
        """ Loads a file into this song object. The new data overwrites any previous
        data stored in this song.

        Args:
            filename (String): Name of the file to load in
            print_file (bool, optional): Whether or not to print out the song. Mainly 
                used for debugging purposes. Defaults to False.
        """
        FileIO.read_midi_file(self, filename=filename, print_file=print_file)
        self.detect_key_and_scale()
        self.get_chord_names()

    def clear_song_data(self):
        """ Deletes all of the data from this song object and resets its default values
        """
        self.tracks = []
        self.ticks_per_beat = DEFAULT_TICKS_PER_BEAT

    def get_c_indexed_note_frequencies(self):
        """ Returns an array containing how many times each of the 12 notes appears in this 
        song, starting with C.

        Returns:
            int[]: note frequencies array
        """
        c_indexed_note_frequency = [0] * 12
        for track in self.tracks:
            if not track.is_percussion:
                track_frequencies = track.get_c_indexed_note_frequencies()
                for idx, val in enumerate(track_frequencies):
                    c_indexed_note_frequency[idx] += val
        return c_indexed_note_frequency

    def get_note_frequencies(self, key):
        """ Returns an array containing how many times each of the 12 notes appears in this 
        song, starting with key.

        Args:
            key (Key): Key to start the array with

        Returns:
            int[]: note frequencies array
        """
        indexed_note_frequency = [0] * 12
        for track in self.tracks:
            if not track.is_percussion:
                track_frequencies = track.get_note_frequencies(key)
                for idx, val in enumerate(track_frequencies):
                    indexed_note_frequency[idx] += val
            return indexed_note_frequency

    def get_tracks_by_tag(self, tag: TagEnum):
        """
        Returns an array of tracks in the song that have the given tag attached
        :param tag: the tag enum that will be searched for
        :return: An array of tracks that match the given tag enum
        """
        tracks = []
        for track in self.tracks:
            if track.tag == tag:
                tracks.append(track)
        return tracks

    def change_song_key_by_half_steps(self, num_half_steps):
        """ Shifts all notes in the song up the by num_half_steps half steps.
        If num_half_steps is negative, the notes will be shifted down instead
        of up.

        Args:
            num_half_steps (int): Number of half steps to move the notes in the song by

        Returns:
            Song: The newly edited song.
        """
        for track in self.tracks:
            if not track.is_percussion:
                for note in track.notes:
                    note.pitch += num_half_steps

        self.get_chord_names()
        return self

    def change_song_key(self, origin_key, destination_key, interval_begin=0, interval_end=float('inf')):
        """ Changes the key of a song for a certain time interval during it.
        TODO: this will not need origin_key once key detection is fully implemented

        Args:
            origin_key (Key): Current key of the song
            destination_key (Key): Key to change the song to
            interval_begin (int): Absolute time during the song to start the key change
            interval_end (int): Absolute time during the song to end the key change

        Raises:
            SyntaxError: If origin_key or destination_key is not a valid Key

        Returns:
            Song: the newly edited song
        """

        # Check to make sure params are the correct type
        if not isinstance(origin_key, Key) or not isinstance(destination_key, Key):
            raise SyntaxError("Parameters are not of the right type.  They must be of type 'Key'")

        # Get the index of the origin key
        origin_index = origin_key.get_c_based_index_of_key()

        # Get the index of the destination key
        destination_index = destination_key.get_c_based_index_of_key()

        # discover offset (this is the number of half steps to move each note to get to the destination key)
        offset = destination_index - origin_index

        origin_steps = []
        dest_steps = []

        # iterate over each scale and find the source and destination scales
        for scale in SCALE_TYPES.items():
            # Break the scale dict into source and destination values
            if scale[0] == str(origin_key.mode):
                origin_steps = scale[1]
            if scale[0] == str(destination_key.mode):
                dest_steps = scale[1]

        # apply the offset to each note within the time interval
        for track in self.tracks:
            if not track.is_percussion:
                for note in track.notes:
                    if interval_begin <= note.time <= interval_end:
                        note.pitch += offset

                        # Reset c_indexed_pitch_class
                        note.c_indexed_pitch_class = note.pitch % 12

                        # Shift each note to its corresponding location in the destination mode
                        origin_pitch = 0
                        dest_pitch = 0
                        for i in range(len(origin_steps)):

                            if (note.c_indexed_pitch_class - origin_key.get_c_based_index_of_key()) % 12 == origin_pitch:
                                note.pitch += (dest_pitch - origin_pitch)
                            # Deal with accidentals here? or leave them where they are?
                            origin_pitch += origin_steps[i]
                            dest_pitch += dest_steps[i]
        return self

    def get_note_velocity_graph(self, name):
        """ Shows a graph of the velocity (intensity/loudness) of all the notes in this song.
        TODO: make this return rather than 'print'
        """
        all_velocity = []
        all_time = []
        for track in self.tracks:
            for note in track.notes:
                all_velocity.append(note.velocity)
                all_time.append(note.time)

        plt.plot(all_time, all_velocity)
        plt.xlabel("Time")
        plt.ylabel("Velocity")
        # Change this to show title of song when that variable is available
        plt.title("Velocity of Notes in " + name)
        plt.show()

    def get_bar_graph(self, title, x_label, y_label, items):
        counter = collections.Counter(items)
        counter = dict(sorted(counter.items(), key=lambda item: item[1], reverse=True))
        bar = plt.bar(x=counter.keys(), height=counter.values())
        plt.xlabel(x_label)
        plt.ylabel(y_label)
        # Change this to show title of song when that variable is available
        plt.title(title)
        return bar

    # Shows a graph of the frequency of all the notes in this song
    def get_note_frequency_graph(self, name):
        """ Shows a graph of the frequency that each note apperas in this song.
        TODO: make this return rather than 'print'
        """
        all_notes = []
        for track in self.tracks:
            for note in track.notes:
                all_notes.append(KEYS[note.pitch % 12])

        graph = self.get_bar_graph("Frequency of Notes in " + name, "Note", "Frequency", all_notes)
        plt.show()

    def to_string(self):
        """ Returns the contents of the song as a string in the format: \n
        Song metadata       \n
        Track1 metadata     \n
        notes               \n
        ...V...             \n
        control messages    \n
        ...V...             \n
        Track 2 metadata    \n
        notes               \n
        ...V...             \n
        etc

        Note: The notes and control messages are listed separately, they are not sorted together by time.
        """
        message = ""
        message += ("Ticks per beat: " + str(self.ticks_per_beat) + "\n")
        for t in self.tracks:
            message += ("  Track name: " + str(t.track_name) + "\n")
            message += ("  Device name: " + str(t.device_name) + "\n")
            for n in t.notes:
                message += ("  Pitch:" + str(n.pitch) + " Velocity: " + str(n.velocity) + " Time: " + str(n.time) +
                            " Duration: " + str(n.duration) + " Channel: " + str(n.channel) + "\n")
            for c in t.controls:
                message += (
                        "  Type: " + str(c.msg_type) + " Tempo: " + str(c.tempo) + " Control: " + str(c.control) +
                        " Value: " + str(c.value) + " Instrument: " + str(c.instrument) + " Time: " + str(c.time) +
                        "\n")
        return message

    def equals(self, song):
        """ Determines whether this song and the specified song are equivalent.

        Args:
            song (Song): Song to check this song against

        Returns:
            bool: True if the songs are equivalent, false otherwise
        """

        if self.ticks_per_beat != song.ticks_per_beat:
            logging.info(msg="These songs have different ticks_per_beat values")
            logging.info(msg="This: " + str(self.ticks_per_beat) + ", compare to: " + str(song.ticks_per_beat))
            return False

        if len(self.tracks) != len(song.tracks):
            logging.info(msg="These songs have a different number of tracks")
            logging.info(msg="This: " + str(len(self.tracks)) + ", compare to: " + str(len(song.tracks)))
            return False

        for i, track in enumerate(self.tracks):
            if not track.equals(song.tracks[i]):
                return False

        return True

    def generate_possible_keys_and_scales(self):
        """ 
        Detect the key of a song using Mr. Dehaan's algorithm.  This algorithm generates the valid notes
        for every key and for every scale and checks the occurrences of the notes in the song against the valid
        key/scale notes.  It then finds how many errors (or misses) occurred.  It then finds the key/scale with the
        lowest number of errors (or the list of key/scale with the same minimum) and returns the result.
        :return: A list of Key objects that have the minimum number of errors [0], and the minimum errors [1]
        """

        note_frequencies = self.get_c_indexed_note_frequencies()

        key_and_scale_error_record = {}

        # Iterate over each key
        for key in KEYS:
            # we have to rotate the frequency array to be 0 indexed at the key.
            # get the index of the key we want the frequencies indexed by
            key_index = Key(key).get_c_based_index_of_key()

            # rotate the array left based on the index to have the 0th index contain the frequency of the tonic or key
            key_indexed_note_frequencies = note_frequencies[key_index:] + note_frequencies[:key_index]

            # iterate over each scale and count errors for each
            for scale in SCALE_TYPES.items():

                # Break the scale dict into key and value
                scale_name = scale[0]
                scale_steps = scale[1]

                # keep an error counter to increment each time a note is not in the key/scale
                errors = 0

                # generate an array with true in the indexes that are in the key/scale and false in the
                # indexes that are notes that are out of the scale  (the array is key indexed at 0)
                accepted_notes = [False] * NUM_NOTES

                # semitones that a note in the scale is away from the key (starts at 0 because regardless of scale the
                # tonic will be in the scale)
                semitones_from_key = 0
                accepted_notes[0] = True

                # iterate through the scale semitones marking notes as valid for the key/scale
                for semitones in scale_steps:
                    semitones_from_key += semitones
                    # % number of notes (12) because some signatures will count semitones back to tonic
                    accepted_notes[semitones_from_key % NUM_NOTES] = True

                # count all of the errors that occur (notes in the song that are not accepted by key/scale)
                for idx in range(NUM_NOTES):
                    if not accepted_notes[idx]:
                        errors += key_indexed_note_frequencies[idx]

                # store errors in a record dictionary
                key_and_scale_error_record[key + ' ' + scale_name] = errors

        # find the key/scales with the minimum values in the error dictionary
        minimum_errors = min(key_and_scale_error_record.values())
        result = [k for k, v in key_and_scale_error_record.items() if v == minimum_errors]

        keys = []
        for key in result:
            keys.append(Key(key.split()[0], key.split()[1]))

        return keys, minimum_errors

    def detect_key_and_scale(self):
        """
        Uses the generate possible keys and scales method to get a list of potential keys, determines which is
        the most likely based on the most common notes in the song.

        :return: The detected key [0], the minimum errors [1], and the confidence level [2]
        """

        note_frequencies = self.get_c_indexed_note_frequencies()
        num_notes = sum(note_frequencies)
        keys, minimum_errors = self.generate_possible_keys_and_scales()


        # now we have the relative major and minors, we can use the note frequencies to differentiate
        # between the two based on the assumption that for most cases the tonic will be played more than
        # other notes.  This lets us differentiate scales with the same notes such as D major and B Minor.

        # get the resulting keys/scales
        relative_major_key_scale = None
        relative_minor_key_scale = None
        for key in keys:
            if key.mode == "major":
                relative_major_key_scale = key
            elif key.mode == "minor":
                relative_minor_key_scale = key

        # get the index of the key in order to find its frequency in the frequency array
        idx_of_major_key = relative_major_key_scale.get_c_based_index_of_key()
        idx_of_minor_key = relative_minor_key_scale.get_c_based_index_of_key()

        # get the frequency of each tonic
        major_frequency = note_frequencies[idx_of_major_key]
        minor_frequency = note_frequencies[idx_of_minor_key]

        # compare and return the most common key scale
        if major_frequency >= minor_frequency:
            detected_return_key = relative_major_key_scale
        else:
            detected_return_key = relative_minor_key_scale

        # determine confidence based on number of 1 - number of errors/ number of notes
        confidence = 0
        if num_notes != 0:
            confidence = 1 - minimum_errors / num_notes

        # set the key of the song
        self.key = detected_return_key

        # return the tuple
        return detected_return_key, minimum_errors, confidence

    def detect_key_by_phrase_endings(self):
        """
        Takes the song object and looks at the notes in the melody and bass tracks, and finds the notes with the longest
        pauses after them (likely the ends of melodic phrases). These notes are narrowed down until the set number of
        notes (as a percentage) are found.
        :return: Three objects in the format [key: Key, message: String, confidence: String]. The message contains
        lots of diagnostic information that explains what's going on behind the scenes, and shows a confidence value.
        Note: If the detected tonic is not a note in the detected scale,
        """
        TIME_INTERVAL_INCREASE = 20
        PERCENTAGE_TO_FIND = 0.01

        total_song_notes = 0
        time_interval = 0
        detected_tonic = ""
        message = ""
        confidence = ""

        for track in self.tracks:
            total_song_notes += len(track.notes)

        total_found_notes = total_song_notes
        # total_found_notes = 0

        # Until you find less than the percentage in PERCENTAGE_TO_FIND
        while total_found_notes > PERCENTAGE_TO_FIND * total_song_notes and total_found_notes > len(self.tracks):

            total_found_notes = 0
            time_interval += TIME_INTERVAL_INCREASE

            c_indexed_total_note_frequency = [0] * NUM_NOTES
            for track in self.tracks:
                if track.tag == TagEnum.MELODY or track.tag == TagEnum.BASS:
                    c_indexed_track_note_frequency = [0] * NUM_NOTES
                    for idx, note in enumerate(track.notes):
                        if idx != len(track.notes):
                            # If this note is the last note of the song, or has a long pause after

                            # if idx == len(track.notes) - 1 \
                            #         or (track.notes[idx].time + track.notes[idx].duration) % \
                            #         (4 * self.ticks_per_beat) < time_interval:
                            if idx == len(track.notes) - 1 or track.notes[idx+1].time - note.time > time_interval:

                                total_found_notes += 1
                                c_indexed_track_note_frequency[note.c_indexed_pitch_class] += 1
                                # print(track.track_name + " time: " + str(track.notes[idx-1].time) + " pitch: " +
                                #       str(track.notes[idx-1].c_indexed_pitch_class) + " ch: " + str(track.channel))
                            # if idx == len(track.notes) - 1:
                                # print("Last note: " + str(note.c_indexed_pitch_class))
                    message += (str(c_indexed_track_note_frequency) + ": " + str(track.tag) + " - " +
                                str(track.track_name) + "\n")
                    for i in range(12):
                        c_indexed_total_note_frequency[i] += c_indexed_track_note_frequency[i]

            message += (str(c_indexed_total_note_frequency) + ": totals"  + "\n")

            max_val = 0
            max_idx = -1
            for i in range(len(c_indexed_total_note_frequency)):
                if max_val < c_indexed_total_note_frequency[i]:
                    max_val = c_indexed_total_note_frequency[i]
                    max_idx = i

            if total_found_notes == 0:
                confidence = 0
            else:
                confidence = str(c_indexed_total_note_frequency[max_idx] / total_found_notes)

            message += ("Detected key: " + KEYS[max_idx] + "\n")
            message += ("Time interval: " + str(time_interval) + "\n")
            message += ("Found notes: " + str(total_found_notes) + "\n")
            message += ("Confidence: " + confidence + "\n")
            message += ("ticks per beat: " + str(self.ticks_per_beat) + "\n\n")

            # The detected tonic of the song (NOT a Key object yet)
            detected_tonic = KEYS[max_idx]

        # Convert detected_key into a Key object with the correct scale
        possible_keys = self.generate_possible_keys_and_scales()[0]
        detected_key = None
        for key in possible_keys:
            if key.tonic == detected_tonic:
                detected_key = key

        # If the detected tonic is not in the list of possible keys, choose the major mode of the detected scale.
        if detected_key is None:
            for key in possible_keys:
                if key.mode == 'major':
                    detected_key = key

        # set the key of the song
        self.key = detected_key

        return [detected_key, message, confidence]

    def get_chord_names(self):
        """
        Sets the name field inside chord based on the notes in the chord.
        """
        # It would be helpful for us here to have an accidental field on a key to know if its sharp or flat
        # That's because then you know whether to use the keys or equivalent keys array
        # I think this should work for all natural keys though
        # This also doesn't account for diminished chords
        major = [1, 4, 5]
        minor = [2, 3, 6]
        # Gets the key of the song, as an integer
        key = self.key.get_c_based_index_of_key()
        # Iterate over every chord in every track
        for track in self.tracks:
            for chord in track.chords:
                # Get the name of the first note in the chord which will
                # commonly be the name of the chord.
                first_note_index = chord.notes[0].c_indexed_pitch_class
                first_note_name = KEYS[chord.notes[0].pitch % NUM_NOTES]
                # The distance between our first note and the key will tell
                # us if its major or minor.
                distance_between_notes = first_note_index - key + 1
                if distance_between_notes in major:
                    if distance_between_notes == 5 and len(chord.notes) == 4:
                        chord.name = first_note_name + " Dominant"
                    else:
                        chord.name = first_note_name + " Major"
                else:
                    chord.name = first_note_name + " Minor"
                if len(chord.notes) == 4:
                    chord.name = chord.name + " Seventh"

    def get_transition_graph(self, name):
        # TODO: Change the way it iterates through tracks so the chords are always connected by the time
        # they occur in the song.
        # Create a directed graph object and set dimensions
        graph = graphviz.Digraph(comment="Chord transitions in Song")
        graph.attr(ranksep='0.01', nodesep='0.1', label=name + " Chord Transitions", labelloc="t")
        chord_set = set()
        edges = []
        all_chords = []
        # Iterate over every track in song
        for track in self.tracks:
            if not track.chords:
                continue
            all_chords += track.chords

        all_chords.sort(key=lambda chord: chord.time)
        # Hold the previous chord
        prev_chord = all_chords[0]
        graph.node(prev_chord.name)
        chord_set.add(prev_chord.name)
        # Iterate over every chord in track
        for i in range(1, len(all_chords)):
            curr = all_chords[i]
            # Only add chords that are not already on graph
            if curr.name not in chord_set:
                graph.node(curr.name)
                chord_set.add(curr.name)
            edges.append([prev_chord.name, curr.name])
            print(prev_chord.name + " -> " + curr.name)
            prev_chord = curr

        counter = collections.Counter(tuple(edge) for edge in edges)
        for edge, count in counter.items():
            # print(str(edge[0]), count)
            graph.edge(str(edge[0]), str(edge[1]), label=str(count))
        graph.view()

    def get_notes_at_time(self, time):
        """ Returns all the notes in a song that are playing at a given time

        Args:
            time (int): Time to check

        Returns:
            Note[]: array of notes that occur at the given time
        """        
        notes = []
        for track in self.tracks:
            for note in track.notes:
                if note.time >= time and note.time + note.duration <= time:
                    notes.append(note)

        return notes



from markov_chain import MarkovChain
import numpy as py
from Track import Track
from Note import Note

class MarkovLibrary:

    def __init__(self):
        self.chains = {}
    
    def add_markov_chain(self, chain):
        self.chains[chain.name] = chain

    def generate_pattern(self, song, num_notes, note_tone_chain=None, note_length_chain=None):
        t = Track()
        song.add_track(t)
        current_note = py.random.choice([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], 1)[0]
        t.add_note(Note(pitch=current_note+35, time=0, duration=100))
        for i in range (1, num_notes):
            if note_tone_chain != None:
                next_note_tone = note_tone_chain.generate_next_note(current_note)
                current_note = next_note_tone
            else:
                next_note_tone = py.random.choice([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], 1)[0]
            next_note_tone += 35 # Bump up three octaves
            if note_length_chain != None:
                next_note_length = note_length_chain.generate_next_note(current_note)
                current_note = next_note_tone
            else:
                next_note_length = py.random.choice([1, 2, 3], 1)[0]*100
            time = i*(py.random.choice([75, 100, 125, 150], 1)[0])
            t.add_note(Note(pitch=next_note_tone, time=time, duration=next_note_length))



import string

MAJOR = "major"
IONIAN = "ionian"
DORIAN = "dorian"
PHRYGIAN = "phrygian"
LYDIAN = "lydian"
MIXOLYDIAN = "mixolydian"
MINOR = "minor"
AOLIAN = "aolian"
LOCRIAN = "locrian"

KEYS = ['C', 'Db', 'D', 'Eb', 'E', 'F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B']
EQUIVALENT_KEYS = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
MODES = [MAJOR, DORIAN, PHRYGIAN, LYDIAN, MIXOLYDIAN, MINOR, LOCRIAN]
EQUIVALENT_MODES = [IONIAN, DORIAN, PHRYGIAN, LYDIAN, MIXOLYDIAN, AOLIAN, LOCRIAN]


class Mode:
    MAJOR = "major"
    IONIAN = "ionian"
    DORIAN = "dorian"
    PHRYGIAN = "phrygian"
    LYDIAN = "lydian"
    MIXOLYDIAN = "mixolydian"
    MINOR = "minor"
    AOLIAN = "aolian"
    LOCRIAN = "locrian"


class Key:

    def __init__(self, tonic='C', mode='major'):
        """ Constructor for the Key class.

        Args:
            tonic (str): Value for the tonic of the key of the song. Defaults to 'C'.
            type (str): Type of key the song is in. Defaults to 'major'.

        Raises:
            SyntaxError: If the key is not within the list of keys and their equivalents
        """        
        if tonic not in KEYS and tonic not in EQUIVALENT_KEYS:
            raise SyntaxError("Key '" + str(tonic) +
                              "' needs to be the key and #/b if necessary. Examples: 'C#', 'Db', 'F' etc")
        elif mode.lower() not in MODES and mode.lower() not in EQUIVALENT_MODES:
            raise SyntaxError("Mode " + str(mode) + " needs to be a valid mode. Examples: major, minor, dorian, etc" )
        else:
            self.tonic = tonic
            self.mode = mode.lower()

    def get_c_based_index_of_key(self):
        """ Takes a key (as a string) and converts it to the index of this key based on the NOTES and EQUIVALENCE
        arrays specified at the top of this file

        Raises:
            SyntaxError: The key is not a valid key within the list of keys and equivalents

        Returns:
            int: index of this key in the list of notes
        """        
        if self.tonic in KEYS:
            index = KEYS.index(self.tonic)
        elif self.tonic in EQUIVALENT_KEYS:
            index = EQUIVALENT_KEYS.index(self.tonic)
        else:
            raise SyntaxError("Key '" + str(self.tonic) +
                              "' needs to be the key and #/b if necessary. Examples: 'C#', 'Db', 'F' etc")
        return index




from Rhythm import *
from Song import Song
from Track import TagEnum


song = Song()

# Load the MIDI file into the song object
song.load(filename="../music samples/Mii Channel.mid")

# Rhythm pattern to apply to the song
pattern = [HALF, DOTTED_SIXTEENTH, EIGHTH, SIXTEENTH, QUARTER_REST]

# Apply the rhythm pattern
apply_rhythm_pattern(song=song, track=song.tracks[1], pattern_array=pattern)

# Adds human aspect to rhythm by slightly offsetting rhythm timings
humanify_rhythm(song=song, track=song.tracks[1], humanify_percent=0.1)

#save to midi file
song.save('../../MIDI Files/Demo Output/rhythm.mid')


from __future__ import division
from enum import Enum
import sys

from Track import Track, TagEnum
from Note import Note, NUM_NOTES
import numpy as py
from Control import Control


class chainType(Enum):
    NOTE = 0
    CHORD = 1


class DynamicMarkovChain:

    def __init__(self, name, chain_type=chainType.NOTE, token_length=1):
        """Constructor for Markov Chains

        Args:
            name (String): Name of the chain
            type (String): Type of chain (TODO: to be replaced with an enumeration)
        """
        self.name = name
        self.token_length = token_length
        self.chain_type = chain_type
        self.probabilities = {}

    def add_song(self, song):
        """Ingests a song and adds to the total dictionary. When all note changes are added,
        divides all elements in the dictionary by the total amount of note changes recorded and
        stores this in self.probabilities, which will give the probability matrix.

        Args:
            song (Song): Song to ingest and generate probabilities from
            num_tokens (int): Tells the amount of tokens for each pattern

        Returns:
            dict(string:[(int, int)]): Dictionary of note tokens as keys and lists of tuples with a note
            and percentage of occurrences
        """
        if self.chain_type is chainType.NOTE:
            return self.add_notes(song)
        else:
            return self.add_chords(song)

    def generate_next_note(self, current_note_token):
        """Given a token of previous notes, this will randomly generate a new note given the percetanges from
            the markov chain object.

            Args:
                current_note_token (String): A string of 'token_length' notes in C-index

            Returns:
                int, String: An int of the next note and a string of the new pattern token.
        """
        orig_token = current_note_token
        # If this token is the end of the song, we need to begin in a new position
        if self.probabilities.get(current_note_token) is None:
            # Let's check if there are any other similar patterns in the song
            for i in range(NUM_NOTES):
                new_pattern = current_note_token[:len(current_note_token) - 2] + str(i)
                if new_pattern in list(self.probabilities.keys()):
                    current_note_token = new_pattern
                    break
            # If not, let's just randomly pick a new position to continue in the chain
            if orig_token == current_note_token:
                current_note_token = py.random.choice(list(self.probabilities.keys()))
        note_follow = self.probabilities.get(current_note_token)
        percentage = []
        # print(note_follow)
        # Create a 1-D array of the percentage of a given note with this pattern token
        for i in range(NUM_NOTES):
            found = False
            for note in note_follow:
                if i == note[0]:
                    percentage.append(note[1])
                    found = True
            if not found:
                percentage.append(0)
        # Pick a random new note with the percentages and return the note and new pattern
        print(percentage)
        next_note = py.random.choice([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], 1, p=percentage)
        return next_note[0], current_note_token

    def generate_next_chord(self, current_chord_token):
        if self.probabilities.get(current_chord_token) is None:
            # Let's just randomly pick a new position to continue in the chain
            current_chord_token = py.random.choice(list(self.probabilities.keys()))
        # Get all the chords that have followed our current token in a list
        chord_follow = self.probabilities.get(current_chord_token)
        percentage = []
        available_chords = []
        # Store the total chords and the percentage chance that they occur
        for chord in chord_follow:
            available_chords.append(chord[0])
            percentage.append(chord[1])
        # Choose a chord with the given percentages and return it
        next_chord = py.random.choice(available_chords, 1, p=percentage)
        return next_chord[0], current_chord_token

    def generate_pattern(self, song, num_notes, instrument=0, arpeggio=False, octave=3):
        """ Given a new song object and the number of notes to generate, this method will load that
            amount of new notes into a new track using the existing markov chain.

            Args:
                song (Song): A brand new Song object
                num_notes (int): The number of notes to generate
                instrument (int): the number you want for instrument

            Returns:
                Track: The generated track
        """
        # We are currently making every note an eighth note
        # and every chord a half not for easier testing
        eighth_note = int(song.ticks_per_beat / 2)
        half_note = int(song.ticks_per_beat * 2)
        t = Track()
        # Start at a random place in the markov chain
        current_token = py.random.choice(list(self.probabilities.keys()))
        if self.chain_type is chainType.NOTE:
            current_token_array = current_token.split()
            # Add the beginning notes
            for i in range(self.token_length):
                t.add_note(Note(pitch=int(current_token_array[i]) + (octave * 12), time=i * eighth_note, duration=eighth_note))
            # Iterate through the rest of the song
            for i in range(self.token_length, num_notes):
                # Keep generating a new note until it is not dissonant with the current notes being played
                while True:
                    new_note_needed = False
                    # Generate new note for song
                    next_note_rtn = self.generate_next_note(current_token)
                    next_note_tone, current_token = next_note_rtn[0], next_note_rtn[1]
                    # Get all the notes that occur at this time to see if there is dissonance
                    curr_notes = song.get_notes_at_time(time=i * eighth_note)
                    for note in curr_notes:
                        if abs(next_note_tone - note.c_indexed_pitch_class) <= 1:
                            # Then the note will be dissonant
                            new_note_needed = True
                            if len(self.probabilities.get(current_token)) == 1:
                                current_token = py.random.choice(list(self.probabilities.keys()))

                    if not new_note_needed:
                        break
                print(next_note_tone)
                # Create the new pattern with the new note
                if self.token_length == 1:
                    current_token = str(next_note_tone)
                else:
                    # We need to change the pattern token by adding the new note
                    previous_pattern = current_token.split()
                    current_token = previous_pattern[1]
                    for j in range(2, self.token_length):
                        current_token += " " + previous_pattern[j]
                    current_token += " " + str(next_note_tone)
                # Adding the specified octave is a way to bump the notes up (default 3) octaves so it sounds better
                next_note_tone += (octave * 12)  # Bump up three octave
                new_note = Note(pitch=next_note_tone, time=i * eighth_note, duration=eighth_note)
                # There is a one in seven chance that we have a rest
                if not py.random.randint(0, 7):
                    new_note.velocity = 0
                t.add_note(new_note)
        # Here we are doing the same thing but for chords
        else:
            # We split the token to get all the individual chords
            current_token_array = current_token.split(',')
            # We add all the initial notes to the track before iterating
            for i in range(self.token_length):
                current_chord = current_token_array[i].split()
                for j in range(len(current_chord)):
                    t.add_note(Note(pitch=int(current_chord[j]) + (octave * 12), time=i * half_note, duration=half_note))
            # Now we can iterate until we have made enough notes
            for i in range(self.token_length, num_notes):
                # Check if there is dissonance in the new chord
                while True:
                    # Get new chord and token
                    new_note_needed = False
                    next_chord_rtn = self.generate_next_chord(current_token)
                    next_chord, current_token = next_chord_rtn[0], next_chord_rtn[1]
                    # Get all notes that occur at this time to see if there is dissonance
                    curr_notes = song.get_notes_at_time(time=i * eighth_note)
                    for note in curr_notes:
                        curr_chord = next_chord.split()
                        for chord_note in curr_chord:
                            if abs(chord_note - note.c_indexed_pitch_class) <= 1:
                                # The chord is dissonant
                                new_note_needed = True
                                if len(self.probabilities.get(current_token_array)) == 1:
                                    current_token_array = py.random.choice(list(self.probabilities.keys()))

                    if not new_note_needed:
                        break
                print(next_chord)
                if self.token_length == 1:
                    current_token = next_chord
                else:
                    # We need to change the token so it has the new chord in it
                    current_token = current_token_array[1]
                    for j in range(2, self.token_length):
                        current_token += "," + current_token_array[j]
                    current_token += "," + next_chord
                next_chord_array = next_chord.split()
                # If the user wants arpeggio then split the chord up by 32nd notes
                offset = int(eighth_note / 4)
                for j in range(len(next_chord_array)):
                    if not arpeggio:
                        t.add_note(Note(pitch=int(next_chord_array[j]) + (octave * 12), time=i * half_note, duration=half_note))
                    else:
                        t.add_note(Note(pitch=int(next_chord_array[j]) + (octave * 12), time=(i * half_note) + offset,
                                        duration=half_note))
                        offset += int(eighth_note / 4)

        t.controls.append(Control(msg_type='program_change', instrument=instrument, time=0))
        return t

    def add_chords(self, song):
        """Given a song object, this will add all the chords to our probability dictionary AKA
        markov chain and create the percentages for them.

        Args:
            song (Song): Song to ingest and generate probabilities from
            num_tokens (int): Tells the amount of tokens for each pattern

        Returns:
            dict(string:[(int, int)]): Dictionary of note tokens as keys and lists of tuples with a note
            and percentage of occurrences
        """
        # Get all notes from song in one list
        all_chords = []
        for track in song.tracks:
            if not track.chords or track.tag == TagEnum.PERCUSSION:
                continue
            all_chords += track.chords

        if not all_chords:
            raise AttributeError('There are no chords in this song.')
        # Get first x notes of song
        previous_pattern = all_chords[0].to_string()
        previous_pattern += ","
        for i in range(1, self.token_length):
            previous_pattern += all_chords[i].to_string()
            previous_pattern += ","
        previous_pattern = previous_pattern[:len(previous_pattern) - 2]
        print(previous_pattern)

        # Create dictionary
        pattern_dict = self.probabilities
        # for the rest of the song, look at next chord
        for i in range(self.token_length, len(all_chords)):
            next_chord = all_chords[i].to_string()
            # Add that and the new note as a key/value entry.
            if pattern_dict.get(previous_pattern) is None:

                pattern_dict[previous_pattern] = [[next_chord, 1]]
            # We need to check if the pattern and chord combo exists already
            else:
                found = False
                for chord in pattern_dict.get(previous_pattern):
                    if chord[0] == next_chord:
                        chord[1] = chord[1] + 1
                        found = True
                        break
                if not found:
                    pattern_dict.get(previous_pattern).append([next_chord, 1])
            # Finally, we create the new token to continue
            previous_pattern = all_chords[i - self.token_length].to_string()
            previous_pattern += ","
            for j in range(i - self.token_length + 1, i):
                previous_pattern += " " + all_chords[j].to_string()
                previous_pattern += ","
            previous_pattern = previous_pattern[:len(previous_pattern) - 2]

        # Value would be a list of lists with note and count [(0, 1), (1, 2)]
        # If the key/value pair already exists, add one to count
        for value in pattern_dict.values():
            total = 0.0
            for chord in value:
                total += chord[1]
            for chord in value:
                chord[1] = (chord[1] / total)

        self.probabilities = pattern_dict
        print(pattern_dict)
        return pattern_dict
        # Key is the same pattern, value is a list of all the percentages. [0, 0.5, 0.05, 0.25, ...]

    def add_notes(self, song):
        """Ingests a song and adds to the total dictionary. When all note changes are added,
        divides all elements in the dictionary by the total amount of note changes recorded and
        stores this in self.probabilities, which will give the probability matrix.

        Args:
            song (Song): Song to ingest and generate probabilities from
            num_tokens (int): Tells the amount of tokens for each pattern

        Returns:
            dict(string:[(int, int)]): Dictionary of note tokens as keys and lists of tuples with a note
            and percentage of occurrences
        """
        # Get all notes from song in one list
        all_notes = []
        for track in song.tracks:
            if not track.notes or track.tag == TagEnum.PERCUSSION:
                continue
            for note in track.notes:
                if note.chord_note is True:
                    continue
                all_notes.append(note)
        all_notes.sort(key=lambda notes: notes.time)
        # Get first x notes of song
        previous_pattern = str(all_notes[0].c_indexed_pitch_class)
        for i in range(1, self.token_length):
            note = all_notes[i]
            previous_pattern += " " + str(note.c_indexed_pitch_class)

        # Create dictionary
        pattern_dict = self.probabilities
        # for the rest of the song, look at next note
        for i in range(self.token_length, len(all_notes)):
            next_note = all_notes[i].c_indexed_pitch_class
            # Add that and the new note as a key/value entry.
            if pattern_dict.get(previous_pattern) is None:

                pattern_dict[previous_pattern] = [[next_note, 1]]
            else:
                found = False
                for note in pattern_dict.get(previous_pattern):
                    if note[0] == next_note:
                        note[1] = note[1] + 1
                        found = True
                        break
                if not found:
                    pattern_dict.get(previous_pattern).append([next_note, 1])

            previous_pattern = str(all_notes[i - self.token_length].c_indexed_pitch_class)
            for j in range(i - self.token_length + 1, i):
                note = all_notes[j]
                previous_pattern += " " + str(note.c_indexed_pitch_class)

        # Value would be a list of lists with note and count [(0, 1), (1, 2)]
        # If the key/value pair already exists, add one to count
        for value in pattern_dict.values():
            total = 0.0
            for note in value:
                total += note[1]
            for note in value:
                note[1] = (note[1] / total)

        self.probabilities = pattern_dict
        print(pattern_dict)
        return pattern_dict
        # Key is the same pattern, value is a list of all the percentages. [0, 0.5, 0.05, 0.25, ...]



import pytest
import FileIO
from Control import Control
from Note import Note
from Song import Song
from Track import Track
from Key import Key
import mido


def test_chord_functionality():
    """
        Tests the functionality of a chord object being properly ingested

        Tests using chords in a c-key
    """
    orig = Song()

    orig.load(filename="test MIDI/C_major_chord_single.mid")
    assert orig.tracks[1].chords[0].notes[0].pitch == 48  # The pitch of a C3 note
    assert orig.tracks[1].chords[0].notes[1].pitch == 52  # The pitch of a E3 note
    assert orig.tracks[1].chords[0].notes[2].pitch == 55  # The pitch of a G3 note


def test_chord_names():
    orig = Song()
    orig.load(filename="test MIDI/C_major_chords.mid")

    assert orig.tracks[1].chords[0].name == "C Major"
    assert orig.tracks[1].chords[1].name == "C Major Seventh"
    assert orig.tracks[1].chords[2].name == "D Minor"
    assert orig.tracks[1].chords[3].name == "D Minor Seventh"



"""
    This is the example for how to use the key detection functionality in Song.  Feel free to edit the
    song which is loaded in on line #12 to try detecting keys/scales for other songs.
"""

from Song import Song, SongLibrary

# Create an empty song object
song = Song()

# Load the MIDI file into the song object
song.load(filename=SongLibrary.SIMON_AND_GARFUNKEL_SCARBOROUGH_FAIR)

# Run the key detection algorithm which returns a key object, the number of errors, and percent confidence as a tuple
detected_key, number_of_errors, percent_confidence = song.detect_key_and_scale()

# Print the result
print(detected_key.tonic, detected_key.mode, number_of_errors, percent_confidence)

# or alternatively song.detect_key returns the resulting keys/scales so if you don't care about the inner
# workings of the detection algorithm just print the result (or use it as a list) as such:
#
# print(detect_key)




import mido
import pytest

import FileIO as FileIO
from Control import Control
from Note import Note
from Song import Song
from Track import Track


def test_read_write_midi_file():
    """
    Test FileIO read_midi_file() and write_midi_file() methods
    These are called from the song.load() and song.save() methods
    """

    type_0_orig = Song()
    type_1_orig = Song()
    # type_2_orig = Song()
    type_0_new = Song()
    type_1_new = Song()
    # type_2_new = Song()

    type_0_orig.load(filename="Resources/My-Name-Is.mid")
    type_1_orig.load(filename="Resources/GoodRiddance(TimeOfYourLife).mid")
    # type_2_orig.load(filename="Resources/")

    type_0_orig.save(filename="Resources/My-Name-Is_Output.mid", print_file=True)
    type_1_orig.save(filename="Resources/GoodRiddance(TimeOfYourLife)_Output.mid", print_file=True)

    type_0_new.load(filename="Resources/My-Name-Is_Output.mid")
    type_1_new.load(filename="Resources/GoodRiddance(TimeOfYourLife)_Output.mid")

    # Check that the two song objects are the same by using the equals method in Song
    assert type_0_orig.equals(type_0_new)
    assert type_1_orig.equals(type_1_new)

    # Check that the original song is what we expect by spot checking some notes
    # and control messages that we expect to see
    assert type_0_orig.tracks[1].notes[0].pitch == 29
    assert type_0_orig.tracks[1].notes[10].pitch == 31
    assert type_0_orig.tracks[1].controls[2].instrument == 33

    assert type_1_orig.tracks[16].notes[0].pitch == 43
    assert type_1_orig.tracks[19].notes[20].pitch == 60
    assert type_1_orig.tracks[16].controls[0].control == 7


def test_order_messages():
    """
    Test FileIO order_messages() method
    """
    track = Track()

    track.channel = 2

    track.notes.append(Note(pitch=100, time=1000, duration=100, velocity=50))
    track.notes.append(Note(pitch=101, time=2000, duration=150, velocity=50))
    track.notes.append(Note(pitch=102, time=2100, duration=100, velocity=50))
    track.notes.append(Note(pitch=103, time=3000, duration=300, velocity=50))
    track.notes.append(Note(pitch=104, time=3100, duration=100, velocity=50))

    track.controls.append(Control(msg_type="program_change", instrument=20, time=2050))

    msgs = FileIO.order_messages(track)

    assert len(msgs) == 11

    assert msgs[0] == mido.Message(type='note_on', channel=2, note=100, velocity=50, time=1000)
    assert msgs[1] == mido.Message(type='note_off', channel=2, note=100, velocity=50, time=100)
    assert msgs[2] == mido.Message(type='note_on', channel=2, note=101, velocity=50, time=900)
    assert msgs[3] == mido.Message(type='program_change', channel=2, program=20, time=50)
    assert msgs[4] == mido.Message(type='note_on', channel=2, note=102, velocity=50, time=50)
    assert msgs[5] == mido.Message(type='note_off', channel=2, note=101, velocity=50, time=50)
    assert msgs[6] == mido.Message(type='note_off', channel=2, note=102, velocity=50, time=50)
    assert msgs[7] == mido.Message(type='note_on', channel=2, note=103, velocity=50, time=800)
    assert msgs[8] == mido.Message(type='note_on', channel=2, note=104, velocity=50, time=100)
    assert msgs[9] == mido.Message(type='note_off', channel=2, note=104, velocity=50, time=100)
    assert msgs[10] == mido.Message(type='note_off', channel=2, note=103, velocity=50, time=100)


def test_handle_note():
    """
        Tests the handle_note() method
    """    
    track = Track()
    current_notes = []
    num_notes_per_channel = [0] * 16
    found_chord = [False] * 16

    note_on_ch0_30 = mido.Message(type='note_on', channel=0, note=30, velocity=100, time=200)
    note_on_ch1_30 = mido.Message(type='note_on', channel=1, note=30, velocity=100, time=200)
    note_on_ch0_40 = mido.Message(type='note_on', channel=0, note=40, velocity=100, time=200)

    note_on_ch0_30_vel0 = mido.Message(type='note_on', channel=0, note=30, velocity=0, time=200)
    note_on_ch1_30_vel0 = mido.Message(type='note_on', channel=1, note=30, velocity=0, time=200)
    note_on_ch0_40_vel0 = mido.Message(type='note_on', channel=0, note=40, velocity=0, time=200)

    note_off_ch0_30 = mido.Message(type='note_off', channel=0, note=30, time=200)
    note_off_ch1_30 = mido.Message(type='note_off', channel=1, note=30, time=200)
    note_off_ch0_40 = mido.Message(type='note_off', channel=0, note=40, time=200)

    # Test two notes starting on the same channel, then ending on the same channel
    FileIO.handle_note(track=track, notes=current_notes, msg=note_on_ch0_30, time=1000,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    FileIO.handle_note(track=track, notes=current_notes, msg=note_on_ch0_40, time=1200,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    assert len(track.notes) == 0
    assert len(current_notes) == 2
    assert num_notes_per_channel[0] == 2

    FileIO.handle_note(track=track, notes=current_notes, msg=note_off_ch0_30, time=1400,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    FileIO.handle_note(track=track, notes=current_notes, msg=note_off_ch0_40, time=1600,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    assert len(track.notes) == 2
    assert len(current_notes) == 0
    assert num_notes_per_channel[0] == 0

    # Same as above, but using note_on_velocity_0 messages instead of note_off messages
    FileIO.handle_note(track=track, notes=current_notes, msg=note_on_ch0_40, time=1800,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    FileIO.handle_note(track=track, notes=current_notes, msg=note_on_ch0_30, time=2000,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    assert len(track.notes) == 2
    assert len(current_notes) == 2
    assert num_notes_per_channel[0] == 2

    FileIO.handle_note(track=track, notes=current_notes, msg=note_on_ch0_30_vel0, time=2200,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    FileIO.handle_note(track=track, notes=current_notes, msg=note_on_ch0_40_vel0, time=2400,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    assert len(track.notes) == 4
    assert len(current_notes) == 0
    assert num_notes_per_channel[0] == 0

    # Test two of the same notes starting on different channels, then ending on different channels
    FileIO.handle_note(track=track, notes=current_notes, msg=note_on_ch1_30, time=2600,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    FileIO.handle_note(track=track, notes=current_notes, msg=note_on_ch0_30, time=2800,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    assert len(track.notes) == 4
    assert len(current_notes) == 2
    assert num_notes_per_channel[0] == 1
    assert num_notes_per_channel[1] == 1

    FileIO.handle_note(track=track, notes=current_notes, msg=note_off_ch1_30, time=2700,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    FileIO.handle_note(track=track, notes=current_notes, msg=note_off_ch0_30, time=2900,
                       num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)
    assert len(track.notes) == 6
    assert len(current_notes) == 0
    assert num_notes_per_channel[0] == 0
    assert num_notes_per_channel[1] == 0

    assert track.notes[0].time == 1000
    assert track.notes[1].time == 1200
    assert track.notes[2].time == 2000
    assert track.notes[3].time == 1800
    assert track.notes[4].time == 2600
    assert track.notes[5].time == 2800

    assert track.notes[0].duration == 400
    assert track.notes[1].duration == 400
    assert track.notes[2].duration == 200
    assert track.notes[3].duration == 600
    assert track.notes[4].duration == 100
    assert track.notes[5].duration == 100


def test_handle_control():
    """
        Tests the handle_control() method
    """    

    track = Track()

    control_control_all_fields = mido.Message(type='control_change', control=3, value=10)
    control_control_min_fields = mido.Message(type='control_change')
    control_tempo_all_fields = mido.MetaMessage(type='set_tempo', tempo=2000)
    control_tempo_min_fields = mido.MetaMessage(type='set_tempo')
    control_program_all_fields = mido.Message(type='program_change', program=20)
    control_program_min_fields = mido.Message(type='program_change')
    control_invalid_type = mido.MetaMessage(type='time_signature')

    FileIO.handle_control(track=track, msg=control_control_all_fields, time=1000)
    FileIO.handle_control(track=track, msg=control_control_min_fields, time=1000)
    FileIO.handle_control(track=track, msg=control_tempo_all_fields, time=1000)
    FileIO.handle_control(track=track, msg=control_tempo_min_fields, time=1000)
    FileIO.handle_control(track=track, msg=control_program_all_fields, time=1000)
    FileIO.handle_control(track=track, msg=control_program_min_fields, time=1000)

    with pytest.raises(TypeError):
        FileIO.handle_control(track=track, msg=control_invalid_type, time=1000)
    with pytest.raises(AttributeError):
        FileIO.handle_control(track=track, msg=control_tempo_all_fields, time=None)
    with pytest.raises(AttributeError):
        FileIO.handle_control(track=track, msg=None, time=0)
    with pytest.raises(AttributeError):
        FileIO.handle_control(track=None, msg=control_tempo_all_fields, time=0)

    ctrl = track.controls[0]
    assert(ctrl.msg_type == 'control_change' and ctrl.tempo is None and ctrl.control == 3 and ctrl.value == 10
           and ctrl.instrument is None and ctrl.time == 1000)

    ctrl = track.controls[1]
    assert (ctrl.msg_type == 'control_change' and ctrl.tempo is None and ctrl.control == 0 and ctrl.value == 0
            and ctrl.instrument is None and ctrl.time == 1000)

    ctrl = track.controls[2]
    assert (ctrl.msg_type == 'set_tempo' and ctrl.tempo == 2000 and ctrl.control is None and ctrl.value is None
            and ctrl.instrument is None and ctrl.time == 1000)

    ctrl = track.controls[3]
    assert (ctrl.msg_type == 'set_tempo' and ctrl.tempo == 500000 and ctrl.control is None and ctrl.value is None
            and ctrl.instrument is None and ctrl.time == 1000)

    ctrl = track.controls[4]
    assert (ctrl.msg_type == 'program_change' and ctrl.tempo is None and ctrl.control is None and ctrl.value is None
            and ctrl.instrument == 20 and ctrl.time == 1000)

    ctrl = track.controls[5]
    assert (ctrl.msg_type == 'program_change' and ctrl.tempo is None and ctrl.control is None and ctrl.value is None
            and ctrl.instrument == 0 and ctrl.time == 1000)

    assert len(track.controls) == 6



class Control:

    def __init__(self, msg_type=None, tempo=None, control=None, value=None, instrument=None, time=None):
        """ Constructor for the Control message class.

        Args:
            msg_type (String): Contains what type of control message the object represents. Currently supported
                messages are: 'program_change', 'set_tempo', and 'control_change'. Defaults to None.

            tempo (int, optional): For "set_tempo" messages. The tempo value (in microseconds per tick). Lower
                values are faster tempos. Defaults to None.

            control (int, optional): For "control_change" type messages. Stores information about what effect will be
                applied to the given track. Defaults to None. (For more information on control values, see:
                https://beatbars.com/blog/what-does-cc-stand-for-in-midi.html )

            value (int, optional): For "control_change" type messages. Represents the amount of the above effect that
                will be applied. Defaults to None.

            instrument (String, optional): For "program_change" type messages. Instrument of the track.
                Defaults to None. (For more information on program change values, see:
                https://beatbars.com/blog/what-is-program-change.html )

            time (int, optional): For all control messages. The absolute time (in ticks) from the start of the
                song where this message occurs. Defaults to None.
        """
        self.msg_type = msg_type
        self.tempo = tempo
        self.control = control
        self.value = value
        self.instrument = instrument
        self.time = time

    def duplicate_control(self):
        return Control(msg_type=self.msg_type, tempo=self.tempo, control=self.control, value=self.value,
                       instrument=self.instrument, time=self.time)



from Rhythm import *
from Song import Song, SongLibrary
from Track import Track
from dynamic_markov_chain import DynamicMarkovChain, chainType

# Create a new song object and read in the desired training song
training_song = Song()
training_song.load(filename=SongLibrary.NIRVANA_LITHIUM)

# Create a chord and note markov chain with the given attributes
chord_chain = DynamicMarkovChain("chord chain", token_length=3, chain_type=chainType.CHORD)
note_chain = DynamicMarkovChain("note chain", token_length=4, chain_type=chainType.NOTE)

# Train the new markov chains with the desired song
chord_chain.add_song(training_song)
note_chain.add_song(training_song)

# Create a new song to write to
output_song = Song()

# Create verse sections from the chord and note chains
chd_verse = chord_chain.generate_pattern(output_song, 4, 46, octave=4)
mel_verse = note_chain.generate_pattern(output_song, 12, 72, octave=5)

# Create chorus sections from the chord and note chains
chd_chorus = chord_chain.generate_pattern(output_song, 3, 46, octave=4)
mel_chorus = note_chain.generate_pattern(output_song, 8, 24, octave=5)
mel_chorus2 = note_chain.generate_pattern(output_song, 8, 24, octave=5)

# Create a counter melody with the note chain
# bass_mel = note_chain.generate_pattern(output_song, 12, 18, octave=4)

# Apply custom rhythms to the generated tracks
Chord_Verse_Rhythm = [HALF, QUARTER, HALF, HALF]
Chord_Chorus_Rhythm = [HALF, WHOLE, HALF]
Melody_Verse_Rhythm = [QUARTER, EIGHTH, EIGHTH, SIXTEENTH, SIXTEENTH, SIXTEENTH, QUARTER, QUARTER, SIXTEENTH, EIGHTH, EIGHTH, QUARTER]
Melody_Chorus_Rhythm = [QUARTER, EIGHTH, EIGHTH, SIXTEENTH_TRIPLET, SIXTEENTH_TRIPLET, SIXTEENTH_TRIPLET, EIGHTH, QUARTER]
# Bass_Melody_Rhythm = [HALF, HALF, WHOLE]

apply_rhythm_pattern(song=output_song, track=chd_verse, pattern_array=Chord_Verse_Rhythm)
apply_rhythm_pattern(song=output_song, track=chd_chorus, pattern_array=Chord_Chorus_Rhythm)
apply_rhythm_pattern(song=output_song, track=mel_verse, pattern_array=Melody_Verse_Rhythm)
apply_rhythm_pattern(song=output_song, track=mel_chorus, pattern_array=Melody_Chorus_Rhythm)
apply_rhythm_pattern(song=output_song, track=mel_chorus2, pattern_array=Melody_Chorus_Rhythm)
# apply_rhythm_pattern(song=output_song, track=bass_mel, pattern_array=Bass_Melody_Rhythm)

# Create output tracks (different to the individual sections created earlier)
melody_track = Track(channel=0)
# bass_track = Track(channel=1)
chord_track = Track(channel=2)

# "Write" the song by stitching together the individual sections into a complete track
chord_track = chord_track.append_tracks([0, 0, 1, 1, 0, 0, 1, 1, 0, 0], [chd_verse, chd_chorus])
melody_track = melody_track.append_tracks([0, 0, 1, 1, 2, 1, 0, 0, 1, 2, 1, 2, 0, 0], [mel_verse, mel_chorus, mel_chorus2])
# bass_track = bass_track.append_tracks([0, 0, 0, 0, 0, 0, 0, 0], [bass_mel])

# Add these tracks to the song object
output_song.add_track(chord_track)
output_song.add_track(melody_track)
# output_song.add_track(bass_track)

# Save the new written song
output_song.save(filename='../../MIDI Files/Demo Output/Full Song.mid')


import collections

from Song import Song
from Track import Track
import FileIO as FileIO
import matplotlib.pyplot as plt


class Genre:

    def __init__(self, songs=None, name=None):
        """ Constructor for the Genre object

        Args:
            songs (Song[], optional): List of Songs in this genre. Defaults to None.
            name (String): Name of this genre. Defaults to None.
        """        
        if songs is None:
            songs = []
        self.songs = songs
        self.name = name

    def add_song(self, song):
        """ Adds a song to the genre's list of songs

        Args:
            song (Song): The song to add to the Genre
        """        
        assert isinstance(song, Song)
        self.songs.append(song)

    def print_songs(self):
        """ Prints the list of songs in this genre
        TODO: make this private
        """        
        print(self.songs)

    def get_notes_frequency_graph(self):
        """ Creates and displays the visualization of the frequency of each note within
        Songs of this genre.
        TODO: make this return rather than 'print'
        """        
        notes_list = Song.get_notes_array()
        all_notes = []
        for song in self.songs:
            for track in song.tracks:
                for note in track.notes:
                    all_notes.append(notes_list[note.pitch % 12])

        bar = song.get_bar_graph(title="Frequency of Notes in " + self.name,
                                 x_label="Note", y_label="Frequency", items=all_notes)
        plt.show()

    def get_chord_frequency_graph(self):
        all_chords = []
        for song in self.songs:
            for track in song.tracks:
                # TODO: Change this to track.chords once we implement chords
                all_chords.append(track.chords)

        bar = song.get_bar_graph(title="Frequency of Notes in " + self.name,
                                 x_label="Note", y_label="Frequency", items=all_chords)
        plt.show(bar)

    # def get_octave_frequency_graph(self):
    #     all_octaves = []
    #     for song in self.songs:
    #         for track in song.tracks:
    #             for note in track.notes:
    #                 all_octaves.append(note.pitch / 12)



# This object represents each note in the song, and stores data about each note
NUM_NOTES = 12
MIDDLE_C = 60
MAX_VELOCITY = 127


class Note:

    def __init__(self, pitch=MIDDLE_C, time=0, duration=1, velocity=MAX_VELOCITY, channel=0, chord_note=False):
        """ Constructor for the Note class.

        Args:
            pitch (int, optional): The pitch of the note represented as an integer. Defaults to MIDDLE_C. (60)
            time (int, optional): The absolute time during the song that the note starts. Defaults to 0.
            duration (int, optional): The duration that the note is held. Defaults to 1.
            velocity (int, optional): The intensity/loudness of the note. Defaults to MAX_VELOCITY (127).
            channel (int, optional): The channel this note was read from. Defaults to 0.
            chord_note (bool, optional): If this note is part of a chord
        """        
        self.pitch = pitch
        self.time = time
        self.duration = duration
        self.velocity = velocity
        self.channel = channel
        self.chord_note = chord_note
        # The pitch class the note belongs to, stored as an int from 0-11, with
        # 0 being C. Octave information is lost in this calculation.
        self.c_indexed_pitch_class = pitch % NUM_NOTES

    def duplicate_note(self):
        return Note(pitch=self.pitch, time=self.time, duration=self.duration, velocity=self.velocity,
                    channel=self.channel, chord_note=self.chord_note)





import pytest

from Control import Control
from Note import Note
from Track import Track, TagEnum


def test_track():
    notes_in = [Note(pitch=60, time=100, duration=100, velocity=100, channel=0, chord_note=False),
                Note(pitch=61, time=200, duration=100, velocity=100, channel=0, chord_note=False),
                Note(pitch=62, time=300, duration=100, velocity=100, channel=0, chord_note=False),
                Note(pitch=63, time=400, duration=100, velocity=100, channel=0, chord_note=False)]

    notes_in2 = [Note(pitch=60, time=100, duration=100, velocity=100, channel=0, chord_note=False),
                 Note(pitch=61, time=200, duration=100, velocity=100, channel=0, chord_note=False),
                 Note(pitch=62, time=300, duration=100, velocity=100, channel=0, chord_note=False),
                 Note(pitch=63, time=400, duration=100, velocity=100, channel=0, chord_note=False)]

    controls_in = [Control(msg_type='control_change', control=3, value=10, time=100),
                   Control(msg_type='program_change', instrument=20, time=200)]

    controls_in2 = [Control(msg_type='control_change', control=3, value=10, time=100),
                    Control(msg_type='program_change', instrument=20, time=200)]

    track = Track(notes=notes_in, controls=controls_in, track_name="test_track_1", device_name="test_device_1",
                  channel=0)

    track2 = Track(notes=notes_in2, controls=controls_in2, track_name="test_track_2", device_name="test_device_2",
                   channel=0)

    notes_expected = [Note(pitch=60, time=100, duration=100, velocity=100, channel=0, chord_note=False),
                      Note(pitch=61, time=200, duration=100, velocity=100, channel=0, chord_note=False),
                      Note(pitch=62, time=300, duration=100, velocity=100, channel=0, chord_note=False),
                      Note(pitch=63, time=400, duration=100, velocity=100, channel=0, chord_note=False),
                      Note(pitch=60, time=600, duration=100, velocity=100, channel=0, chord_note=False),
                      Note(pitch=61, time=700, duration=100, velocity=100, channel=0, chord_note=False),
                      Note(pitch=62, time=800, duration=100, velocity=100, channel=0, chord_note=False),
                      Note(pitch=63, time=900, duration=100, velocity=100, channel=0, chord_note=False)]

    controls_expected = [Control(msg_type='control_change', control=3, value=10, time=100),
                         Control(msg_type='program_change', instrument=20, time=200),
                         Control(msg_type='control_change', control=3, value=10, time=600),
                         Control(msg_type='program_change', instrument=20, time=700)]

    track_expected = Track(notes=notes_expected, controls=controls_expected, track_name="test_track_1",
                           device_name="test_device_1", channel=0)

    assert track.get_c_indexed_note_frequencies() == [1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0]

    track.generate_tags()
    assert track.tag == TagEnum.MELODY

    track_actual = track.append_track(track=track2)

    assert track_actual.equals(track_expected)






import sys

import mido
from mido import MidiFile

from Chord import Chord
from Control import Control
from Track import Track
from Note import Note


def read_midi_file(song, filename, print_file=False):
    """Takes a song and file name as input, clears the song data and
    overwrites it with the data from the new file. This creates a list
    of tracks, gets the metadata for the song, and created a "song" object

    Args:
        song (Song): Song object to store the MIDI file's song in
        filename (String): filename of MIDI file
        print_file (bool, optional): Whether or not to print the MIDI file
            after reading it in. Defaults to False.

    Raises:
        NotImplementedError: If a Type 2 MIDI file is read

    Returns:
        song: The Song object containing the song in the MIDI file
    """
    midi = MidiFile(filename)

    if print_file:
        print_midi(filename=filename, file=midi)

    song.clear_song_data()
    song.ticks_per_beat = midi.ticks_per_beat

    # File has multiple synchronous tracks or one track with multiple channels
    if midi.type == 1 or midi.type == 0:

        # For each mido track in the file
        for read_track in midi.tracks:

            # The name of this track
            track_name = None
            # The name of the device this track is played on
            device_name = None
            # A list of tracks to be added to the song (Will only have one element for type 1 and 2 files)
            tracks = []
            # Notes that have had their note_on message read, but don't yet have a note_off message
            current_notes = []
            # The current running time of the song (In absolute terms)
            current_time = 0
            # Keeps track of number of concurrent notes in a track or channel
            num_notes_per_channel = [0] * 16
            # Keeps track of if a chord has been detected already on a channel
            # (prevents the same chord from being counted multiple times)
            found_chord = [False] * 16
            # For each message in the mido track
            for msg in read_track:

                # This is the track currently being modified
                track = None

                # Add the delay between notes to the current time
                current_time += msg.time

                # Set the current track to the track with the same channel as the current message.
                # If this track does not yet exist, create it.
                if hasattr(msg, 'channel'):
                    track = set_current_track_by_channel(msg=msg, tracks=tracks)

                # If this message is a note and not metadata
                if hasattr(msg, 'note') and hasattr(msg, 'velocity'):
                    handle_note(msg=msg, notes=current_notes, time=current_time, track=track,
                                num_notes_per_channel=num_notes_per_channel, found_chord=found_chord)

                if msg.type == 'control_change' or msg.type == 'program_change' or msg.type == 'set_tempo':
                    if msg.type == 'set_tempo':
                        if len(tracks) > 0:
                            track = tracks[0]  # set_tempo applies to the whole song, regardless of what track it's on
                        else:
                            track = Track(channel=0)  # If there are no tracks yet, create a new one
                            tracks.append(track)

                    handle_control(msg=msg, track=track, time=current_time)

                # If this message is a program change, (tells the instrument being played on this track)
                if msg.type == 'program_change':
                    track.instrument = msg.program

                # If this message is a track name
                if msg.type == 'track_name':
                    track_name = msg.name

                # If this message is a track device
                if msg.type == 'device_name':
                    device_name = msg.name

                # In case there are notes left in the "current_notes" list when the song ends
                if msg.type == 'end_of_track':
                    conv_remaining_notes(current_notes=current_notes, current_time=current_time, tracks=tracks)

            # Add this track and its associated notes to the song (sorted by time)
            for t in tracks:
                t.track_name = str(track_name)
                t.device_name = str(device_name)
                t.notes.sort(key=lambda note: note.time)
                t.generate_tags()
                song.add_track(t)
        return song

    # File has multiple asynchronous tracks
    else:
        raise NotImplementedError("'Type 2' Midi files are not supported at this time")


def conv_remaining_notes(current_notes, current_time, tracks):
    """ In the case that a midi song has note_on messages that do not have a closing note_off message, this method
    will end those notes when the end_of_track message is read in for that particular track. This should not occur,
    but adding this method increases the program's robustness and allows us to fix malformed midi files

    Args:
        current_notes (Note[]): An array of notes for which a note_on message was received, but no note_off message was
        current_time (int): Current absolute time in the song
        tracks (Track[]): An array of track objects (generated from the current midi track being read in)
    """
    for n in current_notes:
        n.duration = current_time - n.time
        if n.duration == 0:
            n.duration = 1  # A note can't have zero duration (This breaks the file output)
        current_notes.remove(n)
        track = set_current_track_by_channel(msg=n, tracks=tracks)
        track.add_note(n)


def set_current_track_by_channel(msg, tracks):
    """ Returns the track that is assigned to the same channel as the given message/note. If this track does not
    exist, create a new track and set it to this channel.

    Args:
        msg (Message or Note): A Mido Message object, or a Note object for which the associated track needs to be found
        tracks (Track[]): An array of possible tracks that could match the channel of the message or note
    """
    track = None
    for t in tracks:
        if t.channel == msg.channel:
            track = t
    if track is None:
        track = Track(channel=msg.channel)
        tracks.append(track)
    return track


def write_midi_file(song, filename, print_file=False):
    """ Takes a file name and a song as input, recreates a midi file
    using the tracks and metadata included in the song object, and
    exports it to a file with the given name

    Args:
        song (Song): Song object containing the song to be output
        filename (String): Name of the file to write to
        print_file (bool, optional): Whether or not to print the midi
            file. Defaults to False.
    """
    # Generate new type 1 midi file with the original song's metadata
    midi = MidiFile(type=1)
    midi.ticks_per_beat = song.ticks_per_beat

    # For each track in the song
    for i, t in enumerate(song.tracks):
        # Get an ordered list of messages
        msgs = order_messages(t)
        # Create a new midi track and add these messages to the track
        midi.add_track(name=None)

        # Set the name, device, and instrument for the new track
        name_msg = mido.MetaMessage(type='track_name', name=str(t.track_name), time=0)
        midi.tracks[i].append(name_msg)
        device_msg = mido.MetaMessage(type='device_name', name=str(t.device_name), time=0)
        midi.tracks[i].append(device_msg)

        for m in msgs:
            midi.tracks[i].append(m)

        # TODO: This change is temporary for instrument change
        # midi.tracks[i].append(mido.Message(type='program_change', channel=0, program=46, time=0))
        # Add an End of Track message. Time is 500 to allow a small buffer zone from the last note
        eot_msg = mido.MetaMessage(type='end_of_track', time=500)
        midi.tracks[i].append(eot_msg)

    # midi.tracks[0].append(mido.Message(type='program_change', channel=0, program=46, time=0))
    # Save this midi file
    midi.save(filename)

    if print_file:
        print_midi(filename, midi)


def order_messages(track):
    """ Takes a track as input and returns an ordered list of midi messages.
    These messages consist of note_on, note_off, and control messages, and will have
    their attributes (channel, pitch, velocity, time, etc) set correctly based
    on the values stored in each note or control object.

    Args:
        track (Track): Track to order the messages of

    Returns:
        String[]: Array of all messages in order
    """
    note_on = []
    note_off = []
    controls = []
    msgs = []
    time = 0

    track.notes.sort(key=lambda x: x.time)

    # Generate three lists (control messages, note_on and note_off) that store the absolute
    # times when each midi event occurs. These lists consist of tuples,
    # [message, int] that stores a message and it's corresponding time (either
    # starting or ending time in the case of notes)
    for n in track.notes:
        note_on.append([n, n.time])
        note_off.append([n, n.time + n.duration])

    for c in track.controls:
        controls.append([c, c.time])

    # Sort note on and off messages so they're in chronological order
    note_on.sort(key=lambda note: note[1])
    note_off.sort(key=lambda note: note[1])
    controls.sort(key=lambda control: control[1])

    # Compare the first element of the three lists. Write which ever one comes
    # first to a midi message, and remove it from its list.
    while len(note_on) > 0 or len(note_off) > 0 or len(controls) > 0:
        # print("on " + str(len(note_on)) + " off " + str(len(note_off)) + " ctrl " + str(len(controls)))
        if len(note_on) > 0:
            next_note_on = note_on[0][0]
            next_note_on_time = note_on[0][1]
        else:
            next_note_on = None
            next_note_on_time = None

        if len(note_off) > 0:
            next_note_off = note_off[0][0]
            next_note_off_time = note_off[0][1]
        else:
            next_note_off = None
            next_note_off_time = None

        if len(controls) > 0:
            next_control = controls[0][0]
            next_control_time = controls[0][1]
        else:
            next_control = None
            next_control_time = None

        # If the next event is a control change
        if next_control_time is not None and \
                (next_note_off_time is None or next_control_time <= next_note_off_time) and \
                (next_note_on_time is None or next_control_time <= next_note_on_time):

            c = next_control
            msg_time = c.time
            msg_type = 'control'

            if c.msg_type == 'set_tempo':
                msgs.append(mido.MetaMessage(type=c.msg_type, tempo=c.tempo,
                                             time=msg_time - time))
            elif c.msg_type == 'control_change':
                msgs.append(mido.Message(type=c.msg_type, channel=track.channel, control=c.control, value=c.value,
                                         time=msg_time - time))
            else:  # implies this message is a program change
                msgs.append(mido.Message(type=c.msg_type, channel=track.channel, program=c.instrument,
                                         time=msg_time - time))

            controls.remove(controls[0])

        # If the next event is a note_on event
        elif len(note_on) > 0 and next_note_on_time < next_note_off_time:
            n = next_note_on
            msg_type = 'note_on'
            msg_time = next_note_on_time
            note_on.remove(note_on[0])
            msgs.append(mido.Message(type=msg_type, channel=track.channel, note=n.pitch, velocity=n.velocity,
                                     time=msg_time - time))

        # If the next event is a note_off event
        else:
            n = next_note_off
            msg_type = 'note_off'
            msg_time = next_note_off_time
            note_off.remove(note_off[0])
            msgs.append(mido.Message(type=msg_type, channel=track.channel, note=n.pitch, velocity=n.velocity,
                                     time=msg_time - time))

        time = msg_time

    return msgs


def handle_note(msg, notes, time, track, num_notes_per_channel, found_chord):
    """ Handles the case where a note message is read in from a midi file
    If this is a note_on message, create a new Note object and store it
    in the notes[] array (for now)
    If this is a note_off message, find the corresponding Note object in
    the notes[] array, set the duration, and add this Note object to the
    track being edited. Remove this note from notes[].

    Args:
        msg (mido.Message): Message being read in
        notes (Note[]): List of notes that have been started but not ended
        time (int): Current timestamp where this note occurs
        track (Track): Track this note will be added to
        num_notes_per_channel (Int[]): An array noting how many notes are currently
            playing on each channel, for chord detection.
        found_chord (Boolean[]): Flags (for each channel) if a chord was found with the current notes
            that are playing to avoid marking duplicates. Resets when a new note starts
    """

    # If this message is the start of a note
    if msg.type == 'note_on' and msg.velocity > 0:
        # Create a new Note object and add it to the array of currently playing notes

        notes.append(Note(pitch=msg.note, time=time, duration=0, velocity=msg.velocity, channel=msg.channel))
        num_notes_per_channel[msg.channel] += 1
        found_chord[msg.channel] = False

    # If this message is the end of a note
    elif msg.type == 'note_off' or msg.velocity == 0:
        # Find a possible chord in this channel first
        if num_notes_per_channel[msg.channel] >= 3 and not found_chord[msg.channel]:
            chord = []
            for n in notes:
                # Add every current playing note to chord object
                if n.channel == msg.channel:
                    chord.append(n)
                    n.chord_note = True
            # Create new chord object and add it to the track list
            track.add_chord(Chord(notes=chord, time=time))
            found_chord[n.channel] = True

        # For each note that is currently playing
        for n in notes:
            # Check if the pitch and channel are the same (locate the correct note)
            if n.pitch == msg.note and n.channel == msg.channel:
                # Set the duration of this note based on current_time - start time, add it to the track
                n.duration = time - n.time
                notes.remove(n)
                num_notes_per_channel[n.channel] -= 1
                track.add_note(n)
                break


def handle_control(msg, track, time):
    """ Handles control messages (control_change/program_change). These
    can occur at any time in a track, but have to be treated slightly
    differently to normal notes

    Args:
        msg (mido.Message or mido.MetaMessage): Message being read in
        track (Track): Track being modified
        time (int): Current time into the song this message appears

    Raises:
        AttributeError: If time is None
        AttributeError: If msg is None
        AttributeError: If track is None
        TypeError: If message type is not control, program, or tempo
    """

    if time is None:
        raise AttributeError("Time cannot be None")
    if msg is None:
        raise AttributeError("msg cannot be None")
    if track is None:
        raise AttributeError("Track cannot be None")

    if msg.type == 'control_change':
        control = Control(msg_type=msg.type, control=msg.control, value=msg.value, time=time)
    elif msg.type == 'program_change':
        control = Control(msg_type=msg.type, instrument=msg.program, time=time)
    elif msg.type == 'set_tempo':
        control = Control(msg_type=msg.type, tempo=msg.tempo, time=time)
    else:
        raise TypeError("message " + str(msg) + "is not a valid control message")

    track.controls.append(control)


def print_midi(filename='output.txt', file=None):
    """ Prints a MIDI file in a readable format to standard output.
    Mainly used for debugging.

    Args:
        filename (str, optional): Name of file to write to. Defaults to 'output.txt'.
        file (MidiFile): The MidiFile you wish to print (mido object). Defaults to None.
    """
    original_stdout = sys.stdout  # Save a reference to the original standard output

    with open(filename + '.txt', 'w') as f:
        sys.stdout = f  # Change the standard output to the file we created.
        print(file)
    sys.stdout = original_stdout  # Reset the standard output to its original value



from __future__ import division
from enum import Enum
import sys

from Track import Track
from Note import Note
import numpy as py


class MarkovChain:

    def __init__(self, name, type):
        """Constructor for Markov Chains

        Args:
            name (String): Name of the chain
            type (String): Type of chain (TODO: to be replaced with an enumeration)
        """        
        assert isinstance(type, Type)
        self.name = name
        self.type = type
        self.totalCt = None
        self.totals = None
        self.probabilities = None

    def add_track(self, track):
        """Ingests a track and adds to the 2d totals array. When all note changes are added,
        divides all elements in the 2d array by the total amount of note changes recorded and
        stores this in self.probabilities, which will give the probability matrix.

        Args:
            track (Track): Track to ingest and generate probabiliites from

        Returns:
            int[][]: Matrix of probabiliites
        """        
        assert isinstance(track, Track)
        if self.type == Type.NOTE_LENGTH:
            if self.probabilities == None:
                self.totalCt = {}
                self.totals = {}
                self.probabilities = {}
            # Go through song and calculate probabilities of each note length based on the previous note length
            for i in range(0, len(track.notes) - 1):
                if self.totalCt.get(track.notes[i].duration) != None:
                    self.totalCt[track.notes[i].duration] += 1
                else:
                    self.totalCt[track.notes[i].duration] = 1
                # Add an occurance for track.note[i] -> track.note[i+1]
                if self.totals.get([track.notes[i].duration, track.notes[i + 1].duration]) != None:
                    self.totals[track.notes[i].duration, track.notes[i + 1].duration] += 1
                else:
                    self.totals[track.notes[i].duration, track.notes[i + 1].duration] = 1
            # self.probabilities = self.totals / totalCt
            for key, val in self.totals: # Probably not workings 
                self.probabilities[key] = val / max(self.totalCt.get(key), 1)
            return self.probabilities
        elif self.type == Type.NOTE_TONE:
            if self.probabilities is None:
                self.totalCt = py.zeros(12)
                self.totals = py.zeros((12, 12))
                self.probabilities = py.zeros((12, 12))
            # Go through song and calculate probabilities of each note tonic based on the previous note tonic
            # Make 2D array of length 12 x 12 for each note
            for i in range(0, len(track.notes) - 1):
                self.totalCt[track.notes[i].c_indexed_pitch_class] += 1
                # Add an occurance for track.note[i] -> track.note[i+1]
                self.totals[track.notes[i].c_indexed_pitch_class, track.notes[i + 1].c_indexed_pitch_class] += 1
            # self.probabilities = self.totals / totalCt
            for idx in range(0, 12):
                if self.totalCt[idx] == 0:
                    self.totalCt[idx] = 1
                self.probabilities[idx] = self.totals[idx] / self.totalCt[idx]
            return self.probabilities
        elif self.type == Type.CHORD_TONE:
            return "Not yet implemented"
        elif self.type == Type.CHORD_LENGTH:
            return "Not yet implemented"

    def generate_next_note(self, current_note):
        if self.type == Type.NOTE_TONE:
            next_note = py.random.choice([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], 1, p=self.probabilities[current_note])
            return next_note[0]
        if self.type == Type.NOTE_LENGTH:
            raise NotImplementedError

class Type(Enum):
    NOTE_TONE = 0
    NOTE_LENGTH = 1
    CHORD_TONE = 2
    CHORD_LENGTH = 3



from Song import Song, SongLibrary

# Create a new song object
song = Song()

# Load the desired song into the "song" object
song.load(filename=SongLibrary.MEGADETH_SYMPHONY_OF_DESTRUCTION)    # Actual Key: E Phrygian

# Run the Detect Key method on the song object, capture the returned object
key, report, confidence = song.detect_key_by_phrase_endings()

# Prints the full report
print(report)

# Prints the detected key
print(key.tonic + " " + key.mode)

# Prints confidence
print("Confidence: " + confidence)





from os import walk, path, makedirs
from Song import Song
import csv


def generate_batch_key_detection_report(top_directory_path=None):
    '''
    This method will generate a CSV report showing the genre, artist, song and detected key
    of each song, as well as number of notes that did not fit the scale of the detected key/scale.

    :param top_directory_path: This is the path containing your tree of midi files.
            it should have a sub structure that looks like: genre/artist/song
    :return: No return, but there should be a reports folder in teh directory
            this method was called from which contains the new CSV report
    '''

    if top_directory_path is None:
        raise SyntaxError("You need to provide the top level directory path.")

    # Generate reports directory if it doesn't already exist
    if not path.exists(path.relpath("reports")):
        makedirs("reports")

    # open the key detection report so that we are able to add rows to it
    with open(path.relpath("reports/key_detection_report.csv"), "w+", newline="") as outputFile:
        writer = csv.writer(outputFile)
        writer.writerow(["Genre", "Artist", "Song", "Detected Tonic", "Detected Mode", "Errors", "% Confidence",
                         "Detected Tonic by Endings", "Detected Mode by Endings", "% Confidence"])

        # Iterate through the file structure converting to songs and adding to csv file
        for (dirpath, dirnames, filenames) in walk(top_directory_path):
            if filenames.__len__() != 0:
                for file in filenames:
                    (fileName, fileExtension) = path.splitext(file)
                    if fileExtension == ".mid" and file is not None:
                        file_path = dirpath + "//" + file
                        song = Song()
                        song.load(file_path)
                        detected_tuple = song.detect_key_and_scale()
                        detect_by_phrase_endings = song.detect_key_by_phrase_endings()
                        detected_key_by_endings = detect_by_phrase_endings[0]
                        confidence = detect_by_phrase_endings[2]
                        (pwd, genre, artist) = dirpath.split("\\")
                        writer.writerow([genre, artist, fileName, detected_tuple[0].tonic, detected_tuple[0].mode,
                                         detected_tuple[1], str(detected_tuple[2] * 100) + "%",
                                         detected_key_by_endings.tonic, detected_key_by_endings.mode, confidence])



import sys
# insert at 1, 0 is the script path (or '' in REPL)
sys.path.insert(1, '/Users/erinlitty/Desktop/CSC492/2021FallTeam17-DeHaan/src')

from Track import Track
from Key import Key, KEYS
from Scale import SCALE_TYPES
from Note import Note, NUM_NOTES
import FileIO as FileIO


class Chord:

    # Constructor, takes all fields as inputs
    def __init__(self, notes=None, name=None, time=0):
        if notes is None:
            notes = []
        self.notes = notes
        notes.sort(key=lambda x: x.pitch)
        self.name = name
        self.time = time

    def to_string(self):
        new_string = str(self.notes[0].c_indexed_pitch_class)
        for i in range(1, len(self.notes)):
            new_string += " " + str(self.notes[i].c_indexed_pitch_class)
        return new_string

    def duplicate_chord(self):
        notes = []
        for note in self.notes:
            notes.append(note.duplicate_note())
        return Chord(notes=notes, name=self.name, time=self.time)

